from abc import ABC, abstractmethod
from dataclasses import dataclass

from psdmpy.models.input.entity import Entities
from psdmpy.models.input.node import Nodes


@dataclass(frozen=True)
class SystemParticipants(Entities, ABC):
    @property
    def node(self):
        return self.data["node"]

    @property
    def q_characteristic(self):
        return self.data["q_characteristic"]

    def insert_node_id_columns(self, nodes: Nodes) -> None:
        index_to_id = nodes.id.to_dict()

        if "node_id" not in self.data.columns:
            self.data.insert(self.data.columns.get_loc("node") + 1, "node_id", None)
        self.data["node_id"] = self.node.map(index_to_id)

    @staticmethod
    def attributes():
        return Entities.attributes() + ["node", "q_characteristics"]


@dataclass(frozen=True)
class SystemParticipantsWithCapacity(SystemParticipants):
    def capacity(self):
        return self.data[self.capacity_attribute()]

    @staticmethod
    @abstractmethod
    def capacity_attribute() -> str:
        pass

    @staticmethod
    def attributes():
        return Entities.attributes() + ["node", "capacity"]



from __future__ import annotations

import math
from dataclasses import dataclass
from typing import List

import numpy as np
import pandas as pd
from pandas import DataFrame, Series

from psdmpy.models.result.entity import ResultEntities
from psdmpy.models.result.grid.node import NodeResult
from psdmpy.models.result.participant.dict import ResultDict


@dataclass(frozen=True)
class ConnectorResult(ResultEntities):
    @property
    def i_a_ang(self) -> Series:
        return self.data["i_a_ang"]

    @property
    def i_a_mag(self) -> Series:
        return self.data["i_a_mag"]

    @property
    def i_a_complex(self) -> Series:
        return self.i_a_mag * np.exp(1j * np.radians(self.i_a_ang))

    @property
    def i_b_ang(self) -> Series:
        return self.data["i_b_ang"]

    @property
    def i_b_mag(self) -> Series:
        return self.data["i_b_mag"]

    @property
    def i_b_complex(self) -> Series:
        return self.i_b_mag * np.exp(1j * np.radians(self.i_b_ang))

    def calc_apparent_power(
        self, node_res: NodeResult, voltage_level_kv: float, node="a"
    ):  # in kVA
        s_rated_fn = (
            lambda i_complex: math.sqrt(3)
            * node_res.v_complex(voltage_level_kv)
            * np.conj(i_complex)
        )
        if node == "a":
            res = s_rated_fn(self.i_a_complex)
        elif node == "b":
            res = s_rated_fn(self.i_b_complex)
        else:
            raise ValueError("Node must be either hv or lv")
        return res

    @staticmethod
    def attributes() -> List[str]:
        return ["i_a_ang", "i_a_mag", "i_b_ang", "i_b_mag"]


@dataclass(frozen=True)
class ConnectorsResult(ResultDict):
    entities: dict[str, ConnectorResult]

    @property
    def i_a_ang(self) -> DataFrame:
        return pd.concat(
            [
                node_res.i_a_ang.rename(node_res.input_model)
                for node_res in self.entities.values()
            ],
            axis=1,
        )

    @property
    def i_a_mag(self) -> DataFrame:
        return pd.concat(
            [
                node_res.i_a_mag.rename(node_res.input_model)
                for node_res in self.entities.values()
            ],
            axis=1,
        )

    @property
    def i_b_ang(self) -> DataFrame:
        return pd.concat(
            [
                node_res.i_b_ang.rename(node_res.input_model)
                for node_res in self.entities.values()
            ],
            axis=1,
        )

    @property
    def i_b_mag(self) -> DataFrame:
        return pd.concat(
            [
                node_res.i_b_mag.rename(node_res.input_model)
                for node_res in self.entities.values()
            ],
            axis=1,
        )



import json
from typing import Optional, Union

import plotly.graph_objs as go
from pandas import Series
from shapely.geometry import LineString

from psdmpy.models.input.container.grid import GridContainer
from psdmpy.plots.common.utils import BLUE, GREEN, GREY, RED, rgb_to_hex


def grid_plot(
    grid: GridContainer,
    node_highlights: Optional[Union[dict[tuple, str], list[str]]] = None,
    line_highlights: Optional[Union[dict[tuple, str], list[str]]] = None,
    highlight_disconnected: Optional[bool] = False,
) -> go.Figure:
    """
    Plots the grid on an OpenStreetMap. Lines that are disconnected due to open switches will be grey.

    ATTENTION:
    We currently consider the node_b of the switches to be the auxiliary switch node.
    This is not enforced within the PSDM so might not work as expected.
    If that is the case the wrong lines might be grey.

    Args:
        grid (GridContainer): Grid to plot.
        node_highlights (Optional): Highlights nodes. Defaults to None.
                                    List of uuids or dict[(r, g, b), str] with colors.
        line_highlights (Optional): Highlights lines. Defaults to None.
                                    List of uuids or dict[(r, g, b), str] with colors.
    Returns:
        Figure: Plotly figure.
    """
    fig = go.Figure()

    # Get disconnected lines via opened switches
    opened_switches = grid.raw_grid.switches.get_opened()

    disconnected_lines = grid.raw_grid.lines.filter_by_nodes(opened_switches.node_b)
    _, connected_lines = grid.raw_grid.lines.subset_split(disconnected_lines.uuid)

    connected_lines.data.apply(
        lambda line: _add_line_trace(fig, line, highlights=line_highlights), axis=1
    )
    disconnected_lines.data.apply(
        lambda line: _add_line_trace(
            fig,
            line,
            highlights=line_highlights,
            is_disconnected=True,
            highlight_disconnected=highlight_disconnected,
        ),
        axis=1,
    )

    _add_node_trace(fig, grid, node_highlights)

    center_lat = grid.raw_grid.nodes.data["latitude"].mean()
    center_lon = grid.raw_grid.nodes.data["longitude"].mean()

    # Dynamically calculate the zoom level
    lat_range = (
        grid.raw_grid.nodes.data["latitude"].max()
        - grid.raw_grid.nodes.data["latitude"].min()
    )
    lon_range = (
        grid.raw_grid.nodes.data["longitude"].max()
        - grid.raw_grid.nodes.data["longitude"].min()
    )

    zoom = 12 - max(lat_range, lon_range)

    fig.update_layout(
        # mapbox = {"zoom"=10},
        showlegend=False,
        mapbox_style="open-street-map",
        margin={"r": 0, "t": 0, "l": 0, "b": 0},
        mapbox=dict(
            center=dict(lat=center_lat, lon=center_lon),
            zoom=zoom,  # Adjust the zoom level as per the calculated heuristic
            style="open-street-map",
        ),
    )

    return fig


def _add_line_trace(
    fig: go.Figure,
    line_data: Series,
    is_disconnected: bool = False,
    highlights: Optional[Union[dict[tuple, str], list[str]]] = None,
    highlight_disconnected: Optional[bool] = False,
):
    lons, lats = _get_lons_lats(line_data.geo_position)
    hover_text = line_data["id"]

    color = GREEN
    if isinstance(highlights, dict):
        for line_color, lines in highlights.items():
            if line_data.name in lines:
                color = line_color
    elif highlights is not None:
        color = RED if line_data.name in highlights else GREEN

    if (highlight_disconnected is False) and is_disconnected:
        color = GREY

    # Add the lines
    fig.add_trace(
        go.Scattermapbox(
            mode="lines",
            lon=lons,
            lat=lats,
            hoverinfo="skip",  # Skip hoverinfo for the lines
            line=dict(color=rgb_to_hex(color)),
        ),
    )

    # Create a LineString object from the line's coordinates
    line = LineString(zip(lons, lats))

    # Calculate the midpoint on the line based on distance
    midpoint = line.interpolate(line.length / 2)

    # Add a transparent marker at the midpoint of the line for hover text
    fig.add_trace(
        go.Scattermapbox(
            mode="markers",
            lon=[midpoint.x],
            lat=[midpoint.y],
            hoverinfo="text",
            hovertext=hover_text,
            marker=dict(size=0, opacity=0, color=rgb_to_hex(color)),
        )
    )


def _add_node_trace(
    fig: go.Figure,
    grid: GridContainer,
    highlights: Optional[Union[dict[tuple, str], list[str]]] = None,
):
    node_hover_data = grid.get_nodal_sp_count_and_power()
    nodes_data = grid.raw_grid.nodes.data

    def to_hover_text(node_data: Series):
        if node_data.name not in node_hover_data:
            raise ValueError(
                f"Node with uuid: {node_data.name} not found in node_hover_data"
            )

        return (
            node_data["id"]
            + "<br>"
            + "<br>".join(
                [
                    f"{key}={value}"
                    for key, value in node_hover_data[node_data.name].items()
                ]
            )
        )

    def _node_trace(data, color):
        text = data.apply(lambda node_data: to_hover_text(node_data), axis=1).to_list()

        fig.add_trace(
            go.Scattermapbox(
                mode="markers",
                lon=data["longitude"],
                lat=data["latitude"],
                hovertext=text,
                hoverinfo="text",
                marker=dict(size=6, color=rgb_to_hex(color)),
                text=text,
            )
        )

    if highlights is not None:
        if isinstance(highlights, dict):
            rmd = []
            for nodes in highlights.values():
                rmd.extend(nodes)
            rmd = nodes_data.drop(rmd)
            _node_trace(rmd, BLUE)

            # plot highlighted nodes second so they are on top
            for color, nodes in highlights.items():
                highlighted_nodes = nodes_data.loc[nodes]
                _node_trace(highlighted_nodes, color)
        elif isinstance(highlights, list):
            rmd = nodes_data.drop(highlights)
            _node_trace(rmd, BLUE)

            # plot highlighted nodes second so they are on top
            highlighted_nodes = nodes_data.loc[highlights]
            _node_trace(highlighted_nodes, RED)
        else:
            raise ValueError(
                "Invalid type for highlights. We expect a list of ids or a dict of colors and ids."
            )
    else:
        _node_trace(nodes_data, BLUE)


def _get_lons_lats(geojson: str):
    coordinates = json.loads(geojson)["coordinates"]
    return list(zip(*coordinates))  # returns lons, lats



import concurrent.futures
from dataclasses import dataclass
from datetime import datetime
from functools import partial
from typing import Optional, Union

import pandas as pd
from pandas import DataFrame, Series

from psdmpy.io.utils import check_filter
from psdmpy.models.enums import EntitiesEnum, SystemParticipantsEnum
from psdmpy.models.input.container.grid import GridContainer
from psdmpy.models.input.container.mixins import ContainerMixin
from psdmpy.models.result.participant.flex_options import FlexOptionsResult
from psdmpy.models.result.participant.pq_dict import (
    PQResultDict,
    PQWithSocResultDict,
)
from psdmpy.models.result.power import PQResult


@dataclass(frozen=True)
class ParticipantsResultContainer(ContainerMixin):
    ems: PQResultDict
    loads: PQResultDict
    fixed_feed_ins: PQResultDict
    pvs: PQResultDict
    wecs: PQResultDict
    storages: PQWithSocResultDict
    evcs: PQResultDict
    evs: PQWithSocResultDict
    hps: PQResultDict
    flex: FlexOptionsResult

    def __len__(self):
        participants = self.to_list(include_empty=False)
        return sum([len(participant) for participant in participants])

    def __getitem__(self, slice_val: slice):
        if not isinstance(slice_val, slice):
            raise ValueError("Only slicing is supported!")
        start, stop, _ = slice_val.start, slice_val.stop, slice_val.step
        if not (isinstance(start, datetime) and isinstance(stop, datetime)):
            raise ValueError("Only datetime slicing is supported")
        return self.filter_for_time_interval(start, stop)

    @property
    def p(self) -> DataFrame:
        p_series = {
            participants.entity_type.value: participants.p_sum()
            for participants in self.to_list(include_flex=False)
        }
        return pd.DataFrame(p_series).sort_index().ffill().fillna(0)

    @property
    def q(self) -> DataFrame:
        q_series = {
            participants.entity_type.value: participants.q_sum()
            for participants in self.to_list(include_flex=False)
        }
        return pd.DataFrame(q_series).sort_index().ffill().fillna(0)

    def p_sum(self) -> Series:
        return self.p.sum(axis=1).rename("p_sum")

    def q_sum(self) -> Series:
        return self.q.sum(axis=1).rename("q_sum")

    def subset(self, uuids):
        return ParticipantsResultContainer(
            self.ems.subset(uuids),
            self.loads.subset(uuids),
            self.fixed_feed_ins.subset(uuids),
            self.pvs.subset(uuids),
            self.wecs.subset(uuids),
            self.storages.subset(uuids),
            self.evcs.subset(uuids),
            self.evs.subset(uuids),
            self.hps.subset(uuids),
            self.flex.subset(uuids),
        )

    def get_participants(self, sp_type: SystemParticipantsEnum):
        if sp_type == SystemParticipantsEnum.ENERGY_MANAGEMENT:
            return self.ems
        elif sp_type == SystemParticipantsEnum.LOAD:
            return self.loads
        elif sp_type == SystemParticipantsEnum.FIXED_FEED_IN:
            return self.fixed_feed_ins
        elif sp_type == SystemParticipantsEnum.PHOTOVOLTAIC_POWER_PLANT:
            return self.pvs
        elif sp_type == SystemParticipantsEnum.WIND_ENERGY_CONVERTER:
            return self.wecs
        elif sp_type == SystemParticipantsEnum.STORAGE:
            return self.storages
        elif sp_type == SystemParticipantsEnum.EV_CHARGING_STATION:
            return self.evcs
        elif sp_type == SystemParticipantsEnum.HEAT_PUMP:
            return self.hps
        else:
            raise ValueError(
                f"No return value for system participant of type: {sp_type}"
            )

    def find_participant_result(self, uuid: str):
        for participants_res in self.to_list(include_flex=False):
            if uuid in participants_res:
                return participants_res.get(uuid)
        return ValueError(f"No participant result with uuid: {uuid}")

    def to_dict(self, include_empty: bool = True) -> dict[EntitiesEnum, PQResultDict]:
        if include_empty:
            return {res.entity_type: res for res in self.to_list()}
        else:
            return {res.entity_type: res for res in self.to_list() if res.entities}

    def energies(self) -> dict[SystemParticipantsEnum, float]:
        return {
            sp_type: res.energy()
            for sp_type, res in self.to_dict(include_empty=False).items()
            if sp_type != SystemParticipantsEnum.FLEX_OPTIONS
        }

    def load_and_generation_energies(self) -> dict[SystemParticipantsEnum, float]:
        return {
            sp_type: res.load_and_generation()
            for sp_type, res in self.to_dict(include_empty=False).items()
        }

    def sum(self) -> PQResult:
        participant_res = []
        for participant in self.to_list(include_em=False, include_flex=False):
            participant_res.append(participant.sum())
        return PQResult.sum(participant_res)

    def to_list(
        self, include_em: bool = True, include_flex=True, include_empty=True
    ) -> list[PQResultDict]:
        optional = []
        if include_em:
            optional.append(self.ems)
        if include_flex:
            optional.append(self.flex)
        required_participants = [
            self.loads,
            self.hps,
            self.fixed_feed_ins,
            self.wecs,
            self.storages,
            self.evcs,
            self.pvs,
        ]
        all_participants = required_participants + optional
        if not include_empty:
            all_participants = [
                participants
                for participants in all_participants
                if participants.entities
            ]
        return all_participants

    def filter_by_date_time(self, time: Union[datetime, list[datetime]]):
        return ParticipantsResultContainer(
            self.ems.filter_by_date_time(time),
            self.loads.filter_by_date_time(time),
            self.fixed_feed_ins.filter_by_date_time(time),
            self.pvs.filter_by_date_time(time),
            self.wecs.filter_by_date_time(time),
            self.storages.filter_by_date_time(time),
            self.evcs.filter_by_date_time(time),
            self.evs.filter_by_date_time(time),
            self.hps.filter_by_date_time(time),
            self.flex.filter_by_date_time(time),
        )

    def filter_for_time_interval(self, start: datetime, end: datetime):
        return ParticipantsResultContainer(
            self.ems.filter_for_time_interval(start, end),
            self.loads.filter_for_time_interval(start, end),
            self.fixed_feed_ins.filter_for_time_interval(start, end),
            self.pvs.filter_for_time_interval(start, end),
            self.wecs.filter_for_time_interval(start, end),
            self.storages.filter_for_time_interval(start, end),
            self.evcs.filter_for_time_interval(start, end),
            self.evs.filter_for_time_interval(start, end),
            self.hps.filter_for_time_interval(start, end),
            self.flex.filter_for_time_interval(start, end),
        )

    @classmethod
    def from_csv(
        cls,
        simulation_data_path: str,
        delimiter: str,
        simulation_end: datetime,
        grid_container: Optional[GridContainer] = None,
        filter_start: Optional[datetime] = None,
        filter_end: Optional[datetime] = None,
    ):
        check_filter(filter_start, filter_end)
        with concurrent.futures.ProcessPoolExecutor() as executor:
            # warning: Breakpoints in the underlying method might not work when started from ipynb
            pa_from_csv_for_participant = partial(
                ParticipantsResultContainer.from_csv_for_participant,
                simulation_data_path,
                delimiter,
                simulation_end,
                grid_container,
            )
            participant_results = executor.map(
                pa_from_csv_for_participant,
                filter(
                    lambda x: x != SystemParticipantsEnum.FLEX_OPTIONS,
                    SystemParticipantsEnum.values(),
                ),
            )
            participant_result_map = {}
            for participant_result in participant_results:
                participant_result_map[
                    participant_result.entity_type
                ] = participant_result

        res = ParticipantsResultContainer(
            loads=participant_result_map[SystemParticipantsEnum.LOAD],
            fixed_feed_ins=participant_result_map[SystemParticipantsEnum.FIXED_FEED_IN],
            pvs=participant_result_map[SystemParticipantsEnum.PHOTOVOLTAIC_POWER_PLANT],
            wecs=participant_result_map[SystemParticipantsEnum.WIND_ENERGY_CONVERTER],
            storages=participant_result_map[SystemParticipantsEnum.STORAGE],
            ems=participant_result_map[SystemParticipantsEnum.ENERGY_MANAGEMENT],
            evcs=participant_result_map[SystemParticipantsEnum.EV_CHARGING_STATION],
            evs=participant_result_map[SystemParticipantsEnum.ELECTRIC_VEHICLE],
            hps=participant_result_map[SystemParticipantsEnum.HEAT_PUMP],
            flex=FlexOptionsResult.from_csv(
                SystemParticipantsEnum.FLEX_OPTIONS,
                simulation_data_path,
                delimiter,
                simulation_end,
            ),
        )
        return (
            res
            if not filter_start
            else res.filter_for_time_interval(filter_start, filter_end)
        )

    @staticmethod
    def from_csv_for_participant(
        simulation_data_path: str,
        delimiter: str,
        simulation_end: datetime,
        grid_container: Optional[GridContainer],
        participant: SystemParticipantsEnum,
    ):
        if grid_container:
            input_entities = grid_container.participants.get_participants(participant)
        else:
            input_entities = None
        if participant.has_soc():
            return PQWithSocResultDict.from_csv(
                participant,
                simulation_data_path,
                delimiter,
                simulation_end,
                input_entities,
            )
        else:
            return PQResultDict.from_csv(
                participant,
                simulation_data_path,
                delimiter,
                simulation_end,
                input_entities=input_entities,
            )

    @classmethod
    def create_empty(cls):
        return ParticipantsResultContainer(
            loads=PQResultDict.create_empty(SystemParticipantsEnum.LOAD),
            fixed_feed_ins=PQResultDict.create_empty(
                SystemParticipantsEnum.FIXED_FEED_IN
            ),
            pvs=PQResultDict.create_empty(
                SystemParticipantsEnum.PHOTOVOLTAIC_POWER_PLANT
            ),
            wecs=PQResultDict.create_empty(
                SystemParticipantsEnum.WIND_ENERGY_CONVERTER
            ),
            storages=PQResultDict.create_empty(SystemParticipantsEnum.STORAGE),
            ems=PQResultDict.create_empty(SystemParticipantsEnum.ENERGY_MANAGEMENT),
            evcs=PQResultDict.create_empty(SystemParticipantsEnum.ELECTRIC_VEHICLE),
            evs=PQResultDict.create_empty(SystemParticipantsEnum.ELECTRIC_VEHICLE),
            hps=PQResultDict.create_empty(SystemParticipantsEnum.HEAT_PUMP),
            flex=PQResultDict.create_empty(SystemParticipantsEnum.FLEX_OPTIONS),
        )



import concurrent
from concurrent.futures import ProcessPoolExecutor
from dataclasses import dataclass
from datetime import datetime
from typing import List, Optional, Tuple, Union

from psdmpy.io.utils import check_filter
from psdmpy.models.enums import RawGridElementsEnum
from psdmpy.models.input.container.grid import GridContainer
from psdmpy.models.input.container.mixins import ContainerMixin
from psdmpy.models.input.container.participants import (
    SystemParticipantsContainer,
)
from psdmpy.models.result.container.grid import GridResultContainer
from psdmpy.models.result.container.participants import (
    ParticipantsResultContainer,
)
from psdmpy.models.result.grid.connector import ConnectorsResult
from psdmpy.models.result.grid.enhanced_node import EnhancedNodesResult
from psdmpy.models.result.grid.node import NodesResult
from psdmpy.models.result.grid.transformer import Transformers2WResult


@dataclass(frozen=True)
class GridWithResults(ContainerMixin):
    grid: GridContainer
    results: GridResultContainer

    def nodal_energies(self) -> dict[str, float]:
        return {uuid: self.nodal_energy(uuid) for uuid in self.grid.raw_grid.nodes.uuid}

    def to_list(self, include_empty: bool = False) -> list:
        elems = [self.grid, self.results]
        return elems if include_empty else [r for r in elems if r]

    def nodal_energy(self, uuid: str) -> float:
        return self.nodal_result(uuid).participants.sum().energy()

    def nodal_results(self) -> dict[str, GridResultContainer]:
        return {
            node_uuid: self.nodal_result(node_uuid)
            for node_uuid in self.grid.node_participants_map.keys()
        }

    def nodal_result(self, node_uuid: str) -> "GridResultContainer":
        node_participants = self.grid.node_participants_map[node_uuid]
        participants_uuids = node_participants.uuids()
        participants = self.results.participants.subset(participants_uuids)
        return GridResultContainer(
            name=node_uuid,
            nodes=NodesResult(
                RawGridElementsEnum.NODE,
                {node_uuid: self.results.nodes.entities[node_uuid]},
            ),
            lines=ConnectorsResult.create_empty(RawGridElementsEnum.LINE),
            transformers_2w=Transformers2WResult.create_empty(
                RawGridElementsEnum.TRANSFORMER_2_W
            ),
            participants=participants,
        )

    def em_results(
        self,
    ) -> List[Tuple[SystemParticipantsContainer, ParticipantsResultContainer]]:
        uuid_to_connected_asset = self.grid.participants.ems.uuid_to_connected_assets()
        return [
            (
                self.grid.participants.subset(connected_assets + [em_uuid]),
                self.results.participants.subset(connected_assets + [em_uuid]),
            )
            for (em_uuid, connected_assets) in uuid_to_connected_asset.items()
        ]

    def build_enhanced_nodes_result(self):
        nodal_results = self.nodal_results()

        with ProcessPoolExecutor() as executor:
            futures = {
                executor.submit(self.calc_pq, uuid, nodal_result)
                for uuid, nodal_result in nodal_results.items()
            }

            nodal_pq = {}
            for future in concurrent.futures.as_completed(futures):
                uuid, pq = future.result()
                nodal_pq[uuid] = pq

        return EnhancedNodesResult.from_nodes_result(self.results.nodes, nodal_pq)

    def find_participant_result_pair(self, uuid: str):
        return self.grid.participants.find_participant(
            uuid
        ), self.results.participants.find_participant_result(uuid)

    def filter_by_date_time(self, time: Union[datetime, list[datetime]]):
        return GridWithResults(
            self.grid.filter_by_date_time(time), self.results.filter_by_date_time(time)
        )

    def filter_for_time_interval(self, start: datetime, end: datetime):
        return GridWithResults(
            self.grid, self.results.filter_for_time_interval(start, end)
        )

    @classmethod
    def from_csv(
        cls,
        name: str,
        grid_path: str,
        grid_delimiter: str,
        result_path: str,
        result_delimiter: str,
        primary_data_delimiter: Optional[str] = None,
        simulation_end: Optional[datetime] = None,
        filter_start: Optional[datetime] = None,
        filter_end: Optional[datetime] = None,
    ) -> "GridWithResults":
        check_filter(filter_start, filter_end)

        if not primary_data_delimiter:
            primary_data_delimiter = grid_delimiter

        grid = GridContainer.from_csv(
            grid_path, grid_delimiter, primary_data_delimiter=primary_data_delimiter
        )

        if not grid:
            raise ValueError(f"Grid is empty. Is the path correct? {grid_path}")

        results = GridResultContainer.from_csv(
            name,
            result_path,
            result_delimiter,
            simulation_end,
            grid_container=grid,
            filter_start=filter_start,
            filter_end=filter_end,
        )

        if not results:
            raise ValueError(f"Results are empty. Is the path correct? {result_path}")

        return (
            GridWithResults(grid, results)
            if not filter_start
            else GridWithResults(grid, results)
        )

    @classmethod
    def create_empty(cls):
        return GridWithResults(
            GridContainer.create_empty(), GridResultContainer.create_empty()
        )

    @staticmethod
    def calc_pq(uuid, nodal_result: GridResultContainer):
        pq = nodal_result.participants.sum()
        return uuid, pq



import concurrent.futures
import copy
import logging
import os
import re
import uuid
from dataclasses import dataclass
from datetime import datetime
from functools import partial
from pathlib import Path
from typing import Union

import pandas as pd
from pandas import Series

from psdmpy.errors import ComparisonError
from psdmpy.io import utils
from psdmpy.io.utils import df_to_csv, to_date_time
from psdmpy.models.enums import SystemParticipantsEnum, TimeSeriesEnum
from psdmpy.models.result.participant.pq_dict import PQResultDict
from psdmpy.models.result.power import PQResult


@dataclass
class PrimaryData:
    # ts_uuid -> ts
    time_series: PQResultDict
    # participant_uuid -> ts_uuid
    participant_mapping: dict[str, str]

    def __eq__(self, other):
        try:
            self.compare(other)
            return True
        except ComparisonError:
            return False

    def __len__(self):
        return len(self.time_series)

    def __contains__(self, uuid):
        return uuid in self.time_series or uuid in self.participant_mapping

    def __getitem__(self, get):
        match get:
            case str():
                if get in self.participant_mapping:
                    return self.time_series[self.participant_mapping[get]]
                elif get in self.time_series:
                    return self.time_series[get]
                else:
                    raise KeyError(
                        f"{get} neither a valid time series nor a participant uuid."
                    )
            case slice():
                start, stop, step = get.start, get.stop, get.step
                if step is not None:
                    logging.warning("Step is not supported for slicing. Ignoring it.")
                if not (isinstance(start, datetime) and isinstance(stop, datetime)):
                    raise ValueError("Only datetime slicing is supported")
                return self.filter_for_time_interval(start, stop)
            case _:
                raise ValueError(
                    "Only get by uuid or datetime slice for filtering is supported."
                )

    @property
    def p(self):
        return self.time_series.p

    @property
    def q(self):
        return self.time_series.q

    def p_sum(self) -> Series:
        return self.time_series.p_sum()

    def q_sum(self):
        return self.time_series.q_sum()

    def sum(self) -> PQResult:
        return self.time_series.sum()

    def get_for_participants(self, participants):
        time_series = []
        for participant in participants:
            if participant in self.participant_mapping:
                time_series.append(self.get_for_participant(participant))
        return time_series

    def filter_by_date_time(self, time: Union[datetime, list[datetime]]):
        """
        Filters the result by the given datetime or list of datetimes.
        :param time: the time or list of times to filter by
        :return: a new result containing only the given time or times
        """
        ts = self.time_series.filter_by_date_time(time)
        return PrimaryData(ts, self.participant_mapping)

    def filter_for_time_interval(self, start: datetime, end: datetime):
        ts = self.time_series.filter_for_time_interval(start, end)
        return PrimaryData(ts, self.participant_mapping)

    def to_csv(self, path: str, mkdirs=True, delimiter=","):
        write_ts = partial(PrimaryData._write_ts_df, path, mkdirs, delimiter)
        with concurrent.futures.ProcessPoolExecutor() as executor:
            executor.map(write_ts, list(self.time_series.entities.values()))

        # write mapping data
        index = [str(uuid.uuid4()) for _ in range(len(self.participant_mapping))]
        mapping_data = pd.DataFrame(
            {
                "participant": self.participant_mapping.keys(),
                "time_series": self.participant_mapping.values(),
            },
            index=index,
        )
        mapping_data.index.name = "uuid"
        df_to_csv(
            mapping_data,
            path,
            "time_series_mapping.csv",
            delimiter=delimiter,
            index_label="uuid",
        )

    @staticmethod
    def _write_ts_df(path: str, mkdirs: bool, delimiter: str, ts: PQResult):
        data = copy.deepcopy(ts.data)
        if not isinstance(ts.entity_type, TimeSeriesEnum):
            raise ValueError(
                f"Expected entity type to be TypeSeriesEnum but is {ts.entity_type}. Can not determine file name."
            )
        ts_name = ts.entity_type.get_csv_input_file_name(ts.input_model)
        data["uuid"] = [str(uuid.uuid4()) for _ in range(len(ts))]
        data["time"] = data.index
        data.set_index("uuid", inplace=True)
        df_to_csv(
            data,
            path,
            ts_name,
            mkdirs=mkdirs,
            delimiter=delimiter,
            index_label="uuid",
        )

    def compare(self, other):
        if not isinstance(other, type(self)):
            raise ComparisonError(
                f"Type of self {type(self)} != type of other {type(other)}"
            )

        errors = []

        # Compare participant mapping
        participant_ts_self = set([(p, t) for p, t in self.participant_mapping.items()])
        participant_ts_other = set(
            [(p, t) for p, t in other.participant_mapping.items()]
        )
        mapping_differences = participant_ts_self.symmetric_difference(
            participant_ts_other
        )

        if mapping_differences:
            errors.append(
                ComparisonError(
                    f"Differences in participant mapping. Following entries not in both dicts: {mapping_differences}"
                )
            )

        # Compare time series
        try:
            self.time_series.compare(other.time_series)
        except ComparisonError as e:
            errors.extend(e.errors)

        if errors:
            raise ComparisonError(
                f"Found Differences in {type(self)} comparison: ", errors=errors
            )

    @classmethod
    def from_csv(cls, path: str | Path, delimiter: str):
        # get all files that start with "its_"
        path = Path(path).resolve()
        ts_files = [file for file in os.listdir(path) if file.startswith("its_p")]

        time_series_dict = {}

        if ts_files:
            ts_mapping = utils.read_csv(str(path), "time_series_mapping.csv", delimiter)

            participant_mapping = (
                ts_mapping[["participant", "time_series"]]
                .set_index("participant")
                .to_dict()["time_series"]
            )

            pa_read_time_series = partial(
                PrimaryData._read_pd_time_series, path, delimiter
            )

            # TODO: Only parallel reading if a lot of ts files
            with concurrent.futures.ProcessPoolExecutor() as executor:
                time_series = executor.map(pa_read_time_series, ts_files)
                for ts in time_series:
                    time_series_dict[ts.name] = ts

            time_series = PQResultDict(
                SystemParticipantsEnum.PRIMARY_DATA, time_series_dict
            )

            return PrimaryData(time_series, participant_mapping)

        else:
            logging.debug(f"No primary data in path {path}")
            return PrimaryData(
                PQResultDict.create_empty(SystemParticipantsEnum.PRIMARY_DATA), dict()
            )

    @staticmethod
    def _read_pd_time_series(dir_path: str, delimiter: str, ts_file: str):
        ts_types = "|".join([e.value for e in TimeSeriesEnum])
        pattern = (
            f"({ts_types})_"
            + r"([0-9a-f]{8}-[0-9a-f]{4}-4[0-9a-f]{3}-[89ab][0-9a-f]{3}-[0-9a-f]{12}).csv"  # uuid4 regex
        )
        match = re.match(pattern, ts_file)
        if match:
            ts_type, ts_uuid = match.groups()
            data = utils.read_csv(str(dir_path), ts_file, delimiter, index_col="uuid")
            data["time"] = data["time"].apply(
                lambda date_string: to_date_time(date_string)
            )
            data = data.set_index("time", drop=True)
            if "q" not in data.columns:
                data["q"] = 0
            return PQResult(TimeSeriesEnum(ts_type), ts_uuid, ts_uuid, data)

        else:
            raise IOError(
                f"Could not read time series with name {ts_file}. Expected format: e.g. its_p_5022a70e-a58f-4bac-b8ec-1c62376c216b.csv"
            )

    @classmethod
    def create_empty(cls):
        return cls(
            PQResultDict.create_empty(SystemParticipantsEnum.PRIMARY_DATA), dict()
        )



from dataclasses import dataclass

from psdmpy.models.enums import SystemParticipantsEnum
from psdmpy.models.input.participant.participant import SystemParticipants


@dataclass(frozen=True)
class FixedFeedIns(SystemParticipants):
    @staticmethod
    def get_enum() -> SystemParticipantsEnum:
        return SystemParticipantsEnum.FIXED_FEED_IN

    @property
    def s_rated(self):
        return self.data["s_rated"]

    @property
    def cos_phi_rated(self):
        return self.data["cos_phi_rated"]



from dataclasses import dataclass

from pandas import DataFrame, Series

from psdmpy.models.enums import EntitiesEnum, ThermalGridElementsEnum
from psdmpy.models.input.entity import Entities


@dataclass(frozen=True)
class ThermalBus(Entities):
    data: DataFrame

    @staticmethod
    def get_enum() -> EntitiesEnum:
        return ThermalGridElementsEnum.THERMAL_BUS

    def node(self) -> Series:
        return NotImplemented



from dataclasses import dataclass

from psdmpy.models.enums import RawGridElementsEnum
from psdmpy.models.input.connector.connector import Connector
from psdmpy.models.input.container.mixins import HasTypeMixin


@dataclass(frozen=True)
class Transformers2W(HasTypeMixin, Connector):
    @property
    def type_id(self):
        return self.data["type_id"]

    @property
    def tap_pos(self):
        return self.data["tap_pos"]

    @property
    def auto_tap(self):
        return self.data["auto_tap"]

    @property
    def r_sc(self):
        return self.data["r_sc"]

    @property
    def x_sc(self):
        return self.data["x_sc"]

    @property
    def g_m(self):
        return self.data["g_m"]

    @property
    def b_m(self):
        return self.data["b_m"]

    @property
    def s_rated(self):
        return self.data["s_rated"]

    @property
    def v_rated_a(self):
        return self.data["v_rated_a"]

    @property
    def v_rated_b(self):
        return self.data["v_rated_b"]

    @property
    def d_v(self):
        return self.data["d_v"]

    @property
    def d_phi(self):
        return self.data["d_phi"]

    @property
    def tap_side(self):
        return self.data["tap_side"]

    @property
    def tap_neutr(self):
        return self.data["tap_neutr"]

    @property
    def tap_min(self):
        return self.data["tap_min"]

    @property
    def tap_max(self):
        return self.data["tap_max"]

    @staticmethod
    def get_enum() -> RawGridElementsEnum:
        return RawGridElementsEnum.TRANSFORMER_2_W

    @staticmethod
    def entity_attributes() -> list[str]:
        return [
            "tap_pos",
            "auto_tap",
        ]

    @staticmethod
    def type_attributes() -> list[str]:
        return HasTypeMixin.type_attributes() + [
            "r_sc",
            "x_sc",
            "g_m",
            "b_m",
            "s_rated",
            "v_rated_a",
            "v_rated_b",
            "d_v",
            "d_phi",
            "tap_side",
            "tap_neutr",
            "tap_min",
            "tap_max",
        ]



import logging
from dataclasses import dataclass
from typing import Union

import pandas as pd

from psdmpy.models.enums import SystemParticipantsEnum
from psdmpy.models.input.container.mixins import ContainerMixin
from psdmpy.models.input.node import Nodes
from psdmpy.models.input.participant.bm import BiomassPlants
from psdmpy.models.input.participant.em import EnergyManagementSystems
from psdmpy.models.input.participant.evcs import EvChargingStations
from psdmpy.models.input.participant.evs import ElectricVehicles
from psdmpy.models.input.participant.fixed_feed_in import FixedFeedIns
from psdmpy.models.input.participant.hp import HeatPumps
from psdmpy.models.input.participant.load import Loads
from psdmpy.models.input.participant.pv import PhotovoltaicPowerPlants
from psdmpy.models.input.participant.storage import Storages
from psdmpy.models.input.participant.wec import WindEnergyConverters


@dataclass(frozen=True)
class SystemParticipantsContainer(ContainerMixin):
    ems: EnergyManagementSystems
    loads: Loads
    fixed_feed_ins: FixedFeedIns
    pvs: PhotovoltaicPowerPlants
    biomass_plants: BiomassPlants
    wecs: WindEnergyConverters
    storages: Storages
    evs: ElectricVehicles
    evcs: EvChargingStations
    hps: HeatPumps

    def to_list(self, include_empty=False):
        participants = [
            self.ems,
            self.loads,
            self.fixed_feed_ins,
            self.pvs,
            self.biomass_plants,
            self.wecs,
            self.storages,
            self.evs,
            self.evcs,
            self.hps,
        ]
        return (
            participants
            if include_empty
            else [p for p in participants if not p.data.empty]
        )

    def build_node_participants_map(
        self, nodes: Union[Nodes, list[str]]
    ) -> dict[str, "SystemParticipantsContainer"]:
        uuids = nodes.uuid if isinstance(nodes, Nodes) else nodes
        return {uuid: self.filter_by_node(uuid) for uuid in uuids}

    def filter_by_node(self, node_uuid: str):
        loads = self.loads.filter_by_nodes(node_uuid)
        fixed_feed_ins = self.fixed_feed_ins.filter_by_nodes(node_uuid)
        pvs = self.pvs.filter_by_nodes(node_uuid)
        biomass_plants = self.biomass_plants.filter_by_nodes(node_uuid)
        wecs = self.wecs.filter_by_nodes(node_uuid)
        storages = self.storages.filter_by_nodes(node_uuid)
        ems = self.ems.filter_by_nodes(node_uuid)
        evs = self.evs.filter_by_nodes(node_uuid)
        evcs = self.evcs.filter_by_nodes(node_uuid)
        hps = self.hps.filter_by_nodes(node_uuid)
        return SystemParticipantsContainer(
            ems,
            loads,
            fixed_feed_ins,
            pvs,
            biomass_plants,
            wecs,
            storages,
            evs,
            evcs,
            hps,
        )

    def get_participants(self, sp_type: SystemParticipantsEnum):
        if sp_type == SystemParticipantsEnum.ENERGY_MANAGEMENT:
            return self.ems
        elif sp_type == SystemParticipantsEnum.LOAD:
            return self.loads
        elif sp_type == SystemParticipantsEnum.BIOMASS_PLANT:
            return self.biomass_plants
        elif sp_type == SystemParticipantsEnum.FIXED_FEED_IN:
            return self.fixed_feed_ins
        elif sp_type == SystemParticipantsEnum.PHOTOVOLTAIC_POWER_PLANT:
            return self.pvs
        elif sp_type == SystemParticipantsEnum.WIND_ENERGY_CONVERTER:
            return self.wecs
        elif sp_type == SystemParticipantsEnum.STORAGE:
            return self.storages
        elif sp_type == SystemParticipantsEnum.EV_CHARGING_STATION:
            return self.evcs
        elif sp_type == SystemParticipantsEnum.ELECTRIC_VEHICLE:
            return self.evs
        elif sp_type == SystemParticipantsEnum.HEAT_PUMP:
            return self.hps
        else:
            logging.error(f"No return value for system participant of type: {sp_type}")
            return None

    def find_participant(self, uuid: str):
        if uuid in self.loads:
            return self.loads.get(uuid)
        elif uuid in self.fixed_feed_ins:
            return self.fixed_feed_ins.get(uuid)
        elif uuid in self.pvs:
            return self.pvs.get(uuid)
        elif uuid in self.wecs:
            return self.wecs.get(uuid)
        elif uuid in self.biomass_plants:
            return self.biomass_plants.get(uuid)
        elif uuid in self.ems:
            return self.ems.get(uuid)
        elif uuid in self.storages:
            return self.storages.get(uuid)
        elif uuid in self.evs:
            return self.evs.get(uuid)
        elif uuid in self.evcs:
            return self.evcs.get(uuid)
        elif uuid in self.hps:
            return self.hps.get(uuid)
        else:
            raise ValueError(
                f"Particpant with uuid {uuid} could not be found among participants."
            )

    def uuids(self):
        return pd.concat(
            [
                participants.uuid.to_series()
                for participants in self.to_list(include_empty=True)
            ]
        )

    def subset(self, uuids):
        return SystemParticipantsContainer(
            self.ems.subset(uuids),
            self.loads.subset(uuids),
            self.fixed_feed_ins.subset(uuids),
            self.pvs.subset(uuids),
            self.biomass_plants.subset(uuids),
            self.wecs.subset(uuids),
            self.storages.subset(uuids),
            self.evs.subset(uuids),
            self.evcs.subset(uuids),
            self.hps.subset(uuids),
        )

    @staticmethod
    def from_csv(path: str, delimiter: str) -> "SystemParticipantsContainer":
        loads = Loads.from_csv(path, delimiter)
        fixed_feed_ins = FixedFeedIns.from_csv(path, delimiter)
        pvs = PhotovoltaicPowerPlants.from_csv(path, delimiter)
        biomass_plants = BiomassPlants.from_csv(path, delimiter)
        wecs = WindEnergyConverters.from_csv(path, delimiter)
        storages = Storages.from_csv(path, delimiter)
        ems = EnergyManagementSystems.from_csv(path, delimiter)
        evs = ElectricVehicles.from_csv(path, delimiter)
        evcs = EvChargingStations.from_csv(path, delimiter)
        hps = HeatPumps.from_csv(path, delimiter)
        return SystemParticipantsContainer(
            ems,
            loads,
            fixed_feed_ins,
            pvs,
            biomass_plants,
            wecs,
            storages,
            evs,
            evcs,
            hps,
        )

    @classmethod
    def create_empty(cls):
        return cls(
            EnergyManagementSystems.create_empty(),
            Loads.create_empty(),
            FixedFeedIns.create_empty(),
            PhotovoltaicPowerPlants.create_empty(),
            BiomassPlants.create_empty(),
            WindEnergyConverters.create_empty(),
            Storages.create_empty(),
            ElectricVehicles.create_empty(),
            EvChargingStations.create_empty(),
            HeatPumps.create_empty(),
        )



import dataclasses
import os
import re
from dataclasses import dataclass
from functools import reduce
from pathlib import Path
from typing import List, Tuple, Union

import numpy as np
import pandas as pd
from pandas import Series

from psdmpy.models.enums import EntitiesEnum, SystemParticipantsEnum
from psdmpy.models.result.entity import ResultEntities
from psdmpy.processing.dataframe import divide_positive_negative
from psdmpy.processing.series import (
    add_series,
    duration_weighted_sum,
    hourly_mean_resample,
    load_and_generation,
)


@dataclass(frozen=True)
class PQResult(ResultEntities):
    def __add__(self, other: "PQResult"):
        p_sum = add_series(self.p, other.p, "p")
        q_sum = add_series(self.q, other.q, "q")
        summed_data = p_sum.to_frame().join(q_sum)
        return PQResult(
            self.entity_type
            if self.entity_type == other.entity_type
            else SystemParticipantsEnum.PARTICIPANTS_SUM,
            "PQResult - Sum",
            "PQResult - Sum",
            summed_data,
        )

    def __sub__(self, other: "PQResult"):
        return self + (other * -1)

    def __mul__(self, other: Union[float, int]):
        if isinstance(other, float) or isinstance(other, int):
            updated_data = self.data * other
            return PQResult(self.entity_type, self.input_model, self.name, updated_data)
        else:
            raise ValueError(
                f"Multiplication with type {type(other)} not or not yet supported"
            )

    # commutative multiplication (a*b=b*a)
    __rmul__ = __mul__

    def __eq__(self, other):
        if not isinstance(other, PQResult):
            return False
        return (self.data == other.data).all()

    @property
    def p(self):
        return self.data["p"]

    @property
    def q(self):
        return self.data["q"]

    def complex_power(self):
        return self.p + 1j * self.q

    def angle(self):
        return self.complex_power().apply(lambda x: np.angle(x, deg=True))

    def energy(self) -> float:
        return duration_weighted_sum(self.p)

    def load_and_generation(self) -> Tuple[float, float]:
        return load_and_generation(self.p)

    def divide_load_generation(self):
        load, generation = divide_positive_negative(self.data)
        return dataclasses.replace(self, data=load), dataclasses.replace(
            self, data=generation
        )

    def daily_usage(self, device_power_mw):
        return (
            self.p()
            .groupby(self.p.index.date)
            .apply(lambda series: duration_weighted_sum(series.abs()))
            / device_power_mw
        )

    def annual_duration_series(self, drop_index=True):
        return (
            hourly_mean_resample(self.p)
            .sort_values(ascending=False)
            .reset_index(drop=drop_index)
        )

    def hourly_resample(self):
        updated_data = self.data.apply(lambda col: hourly_mean_resample(col), axis=0)
        return PQResult(self.entity_type, self.input_model, self.name, updated_data)

    def to_csv(self, path: str, file_name: str = None, delimiter: str = ","):
        if not isinstance(path, str):
            path = str(path)
        file_name = file_name if file_name else self.get_default_output_name()
        self.data.to_csv(
            os.path.join(path, file_name), sep=delimiter, index_label="time"
        )

    def get_default_output_name(self):
        return self.input_model + "_" + self.entity_type.get_csv_result_file_name()

    @classmethod
    def from_csv(cls, file_path: str | Path, sp_type: EntitiesEnum, name: str = None):
        file_path = Path(file_path).resolve()
        data = pd.read_csv(file_path)
        data["time"] = pd.to_datetime(data["time"])
        data = data.set_index("time", drop=True)
        file_name = os.path.basename(file_path)
        is_default_file_name = re.match(
            f"{sp_type.get_csv_result_file_name()}_.*", file_name
        )
        input_model = file_name.split("_")[-1] if is_default_file_name else file_name
        return cls(sp_type, name if name else input_model, input_model, data)

    @classmethod
    # todo: find a way for parallel calculation
    def sum(cls, results: List["PQResult"]) -> "PQResult":
        if len(results) == 0:
            return PQResult.create_empty(
                SystemParticipantsEnum.PARTICIPANTS_SUM, "", ""
            )
        if len(results) == 1:
            return results[0]
        return reduce(lambda a, b: a + b, results)

    @staticmethod
    def attributes() -> List[str]:
        return ["p", "q"]


@dataclass(frozen=True)
class PQWithSocResult(PQResult):
    def add_with_soc(
        self, this_capacity: float, other: "PQWithSocResult", other_capacity: float
    ):
        p_sum = add_series(self.p, other.p, "p")
        q_sum = add_series(self.q, other.q, "q")
        total_capacity = this_capacity + other_capacity
        soc = add_series(
            self.soc * this_capacity / total_capacity,
            other.soc * other_capacity / total_capacity,
            "soc",
        )
        summed_data = p_sum.to_frame().join([q_sum, soc])
        return PQWithSocResult(
            self.entity_type,
            "PQResult - Sum",
            "PQResult - Sum",
            summed_data,
        )

    @staticmethod
    # todo: find a way for parallel calculation
    def sum_with_soc(
        results: list[Tuple[float, "PQWithSocResult"]]
    ) -> "PQWithSocResult":
        if len(results) == 0:
            return PQWithSocResult.create_empty(
                SystemParticipantsEnum.PARTICIPANTS_SUM, "", ""
            )
        if len(results) == 1:
            return results[0][1]
        this_capacity, agg = results[0]
        for other_capacity, result in results[1::]:
            agg = agg.add_with_soc(this_capacity, result, other_capacity)
        return agg

    @property
    def soc(self) -> Series:
        return self.data["soc"]

    @staticmethod
    def attributes() -> List[str]:
        return ["p", "q", "capacity"]



from dataclasses import dataclass
from typing import List

from pandas import Series

from psdmpy.models.enums import EntitiesEnum, ThermalGridElementsEnum
from psdmpy.models.input.entity import Entities


@dataclass(frozen=True)
class ThermalHouse(Entities):
    @property
    def eth_losses(self):
        return self.data["eth_losses"]

    @property
    def eth_capa(self):
        return self.data["eth_capa"]

    @property
    def target_temperature(self):
        return self.data["target_temperature"]

    @property
    def upper_temperature_limit(self):
        return self.data["upper_temperature_limit"]

    @property
    def lower_temperature_limit(self):
        return self.data["lower_temperature_limit"]

    @staticmethod
    def attributes() -> List[str]:
        return Entities.attributes() + [
            "thermal_bus",
            "eth_losses",
            "eth_capa",
            "target_temperature",
            "upper_temperature_limit",
            "lower_temperature_limit",
        ]

    @staticmethod
    def get_enum() -> EntitiesEnum:
        return ThermalGridElementsEnum.THERMAL_HOUSE

    def node(self) -> Series:
        return NotImplemented



import json
import re

import pandas as pd

from psdmpy.models.input.container.grid import GridContainer


def create_data(data_dict, data_func):
    nr_items = {len(values) for values in data_dict.values()}
    if not len(nr_items) == 1:
        raise ValueError(f"Number of items in each list must be equal. Got {nr_items}.")
    return pd.concat(
        [
            data_func(**dict(zip(data_dict.keys(), values)))
            for values in zip(*data_dict.values())
        ],
        axis=1,
    ).T


def generate_function_from_attributes(cls, file_path):
    """Generate function strings for each subclass and write them to a file."""
    subclass_function_strings = _generate_function_from_attributes(cls)
    with open(file_path, "w") as file:
        file.write("import pandas as pd\n\n\n")
        for classname, function_string in subclass_function_strings.items():
            file.write(function_string + "\n\n")  # Two newlines for separation


def _generate_function_from_attributes(cls):
    """Generate create_data functions based on type attributes."""

    def get_all_subclasses(cls):
        """Recursively get all subclasses of a given class"""
        subclasses = cls.__subclasses__()
        for subclass in list(
            subclasses
        ):  # Use list to create a copy since we'll be modifying the subclasses list
            subclasses.extend(get_all_subclasses(subclass))
        return subclasses

    def camel_to_snake(s: str) -> str:
        s1 = re.sub(
            "(.)([A-Z][a-z]+)", r"\1_\2", s
        )  # Identify words in CamelCase and add underscore
        return re.sub(
            "([a-z0-9])([A-Z])", r"\1_\2", s1
        ).lower()  # Convert the string to lowercase

    subclasses = get_all_subclasses(cls)
    function_strings = {}

    for subclass in subclasses:
        attributes = subclass.attributes()
        # non_default = attributes - default_attributes.keys()
        # default = default_attributes - attributes.keys()
        args_str = ",\n\t".join(attributes)  # Formatting each argument on a new line
        series_items_str = ",\n\t\t".join(
            [f'"{attr}": {attr}' for attr in attributes]
        )  # Formatting items in pd.Series
        function_str = f"""def create_{camel_to_snake(subclass.__name__)}_data(
\t{args_str}
):
\treturn pd.Series({{
\t\t{series_items_str}
\t}})"""
        function_strings[subclass.__name__] = function_str

    return function_strings


def update_line_geo_pos_from_coords(grid: GridContainer):
    nodes = grid.raw_grid.nodes

    def get_coord(uuid):
        lat = nodes.latitude[uuid]
        lon = nodes.longitude[uuid]
        return [lon, lat]

    for uuid, line in grid.raw_grid.lines.data.iterrows():
        geo_pos = line.geo_position
        node_a_uuid = line.node_a
        node_b_uuid = line.node_b
        geo_json = json.loads(geo_pos)
        geo_json["coordinates"] = [get_coord(node_a_uuid), get_coord(node_b_uuid)]
        grid.raw_grid.lines.data.loc[uuid, "geo_position"] = str(json.dumps(geo_json))



from dataclasses import dataclass
from typing import Union

from networkx import Graph

from psdmpy.models.input.connector.lines import Lines
from psdmpy.models.input.connector.switches import Switches
from psdmpy.models.input.connector.transformer import Transformers2W
from psdmpy.models.input.container.mixins import ContainerMixin
from psdmpy.models.input.entity import Entities
from psdmpy.models.input.node import Nodes


@dataclass(frozen=True)
class RawGridContainer(ContainerMixin):
    nodes: Nodes
    lines: Lines
    transformers_2_w: Transformers2W
    switches: Switches

    def to_list(self, include_empty: bool = False) -> list[Entities]:
        grid_elements = [self.nodes, self.lines, self.transformers_2_w, self.switches]
        return grid_elements if include_empty else [e for e in grid_elements if e]

    def get_branches(self, as_graphs=False) -> Union[list[list[str]], list[Graph]]:
        """
        Returns all branches, branching off from the slack node of the grid.
        The branches are either returned as a list of lists, where each list contains the node uuids of the branch,
        starting at the slack node, or rather as subgraphs containing all nodes and edges of the branch.
        Currently only works for single slack node and single voltage level grids.
        Args:
            as_graphs: If True, returns the branches as subgraphs, otherwise as lists of node uuids.
        Returns:
            A list of lists or a list of subgraphs, containing the branches of the grid.
        """

        slack_node = self.nodes.get_slack_nodes()
        if len(slack_node.data) != 1:
            raise ValueError("Currently only implemented for singular slack nodes.")
        transformers = self.transformers_2_w
        slack_transformers = Transformers2W(
            transformers.data[
                (transformers.node_a.isin(slack_node.uuid.to_list()))
                | (transformers.node_b.isin(slack_node.uuid.to_list()))
            ]
        )
        slack_connected_node = (
            set(slack_transformers.node_a)
            .union(slack_transformers.node_b)
            .difference(slack_node.uuid)
        )
        if len(slack_connected_node) > 1:
            raise ValueError(
                "There are multiple nodes connected to the slack node via a transformer."
            )
        elif len(slack_connected_node) == 0:
            raise ValueError("Did not find a slack node!")
        slack_connected_node = slack_connected_node.pop()
        graph = self.build_networkx_graph()
        branches = self._find_branches(graph, slack_connected_node)
        branches = [[slack_connected_node] + branch for branch in branches]
        if as_graphs:
            return [graph.subgraph(branch).copy() for branch in branches]
        return branches

    def build_networkx_graph(self, include_transformer: bool = False) -> Graph:
        graph = Graph()
        closed_switches = self.switches.get_closed()
        line_data_dicts = self.lines.data.apply(
            lambda row: {"length": row["length"]}, axis=1
        )

        graph.add_nodes_from(self.nodes.uuid)
        graph.add_edges_from(zip(self.lines.node_a, self.lines.node_b, line_data_dicts))
        graph.add_edges_from(zip(closed_switches.node_a, closed_switches.node_b))
        if include_transformer:
            graph.add_edges_from(
                zip(self.transformers_2_w.node_a, self.transformers_2_w.node_b)
            )
        return graph

    @staticmethod
    def _find_branches(G: Graph, start_node):
        visited = set()
        visited.add(start_node)
        branches = []

        def dfs(node, path):
            visited.add(node)
            path.append(node)

            for neighbor in G.neighbors(node):
                if neighbor not in visited:
                    dfs(neighbor, path)

        for neighbor in G.neighbors(start_node):
            if neighbor not in visited:
                path = []
                dfs(neighbor, path)
                branches.append(path)

        return branches

    @classmethod
    def from_csv(cls, path: str, delimiter: str) -> "RawGridContainer":
        nodes = Nodes.from_csv(path, delimiter)
        lines = Lines.from_csv(path, delimiter)
        transformers_2_w = Transformers2W.from_csv(path, delimiter)
        switches = Switches.from_csv(path, delimiter)
        return cls(
            nodes=nodes,
            lines=lines,
            transformers_2_w=transformers_2_w,
            switches=switches,
        )

    @classmethod
    def create_empty(cls):
        return cls(
            nodes=Nodes.create_empty(),
            lines=Lines.create_empty(),
            transformers_2_w=Transformers2W.create_empty(),
            switches=Switches.create_empty(),
        )



import logging
import os
from abc import ABC, abstractmethod

from psdmpy.errors import ComparisonError
from psdmpy.io.utils import df_to_csv
from psdmpy.models.enums import EntitiesEnum
from psdmpy.models.input.connector.connector import Connector
from psdmpy.models.input.entity import Entities
from psdmpy.models.input.participant.participant import SystemParticipants


class ContainerMixin(ABC):
    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False
        for a, b in zip(
            self.to_list(include_empty=True), other.to_list(include_empty=True)
        ):
            if a != b:
                return False
        return True

    def __bool__(self):
        return len(self.to_list(include_empty=False)) > 0

    @abstractmethod
    def to_list(self, include_empty: bool = False) -> list:
        pass

    @classmethod
    @abstractmethod
    def create_empty(cls):
        pass

    def to_csv(self, path: str, mkdirs=True, delimiter: str = ","):
        for entities in self.to_list():
            try:
                entities.to_csv(path, delimiter=delimiter, mkdirs=mkdirs)
            except Exception as e:
                logging.error(f"Could not write {type(entities)} to {path}. Error: {e}")

    def compare(self, other) -> None:
        """
        Compares the grid with another grid, and raises an error if they are not equal.
        """
        if not isinstance(other, type(self)):
            raise ComparisonError(
                f"Type of self {type(self)} != type of other {type(other)}"
            )
        errors = []

        list_a = self.to_list(include_empty=True)
        types_a = [type(e) for e in list_a]
        list_b = other.to_list(include_empty=True)
        types_b = [type(e) for e in list_b]

        if types_a != types_b:
            raise ComparisonError(
                f"Non empty container elements differ: {types_a} != {types_b}"
            )

        for a, b in zip(list_a, list_b):
            try:
                a.compare(b)
            except ComparisonError as e:
                errors += [e]

        if errors:
            raise ComparisonError(f"Comparison of {type(self)} failed.", errors=errors)


class HasTypeMixin(ABC):
    """
    Mixin for entities that have a separate type within the psdm data model.
    This should only be extended as an addition to the Entities class.
    """

    @property
    def type_id(self):
        return self.data["type_id"]

    @property
    def type_uuid(self):
        return self.data["type_uuid"]

    def to_csv(self, path: str, mkdirs=True, delimiter: str = ","):
        if mkdirs:
            os.makedirs(path, exist_ok=True)
        # persist entity_input.csv
        all_entity_attributes = self.attributes(include_type_attrs=False)
        all_entity_attributes.append("type_uuid")
        entity_data = self.data[all_entity_attributes].rename(
            columns={"type_uuid": "type"}
        )
        df_to_csv(
            entity_data,
            path,
            self.get_enum().get_csv_input_file_name(),
            mkdirs=mkdirs,
            delimiter=delimiter,
        )

        # persist entity_type_input.csv
        type_data = (
            self.data[self.type_attributes()]
            .set_index("type_uuid", drop=True)
            .rename(columns={"type_id": "id"})
        )
        df_to_csv(
            type_data.drop_duplicates(),
            path,
            self.get_enum().get_type_file_name(),
            mkdirs=mkdirs,
            delimiter=delimiter,
        )

    @classmethod
    def attributes(cls, include_type_attrs: bool = True) -> list[str]:
        base_attributes = []
        # check if cls is a subclass of SystemParticipants
        if issubclass(cls, SystemParticipants):
            base_attributes += SystemParticipants.attributes()
        elif issubclass(cls, Connector):
            base_attributes += Connector.attributes()
        elif issubclass(cls, Entities):
            base_attributes += Entities.attributes()
        else:
            logging.warning(
                "We expect the type that extends this mixin to be a subclass of a type with attributes!"
            )
        all_entity_attributes = base_attributes + cls.entity_attributes()
        return (
            all_entity_attributes + cls.type_attributes()
            if include_type_attrs
            else all_entity_attributes
        )

    @staticmethod
    @abstractmethod
    def entity_attributes() -> list[str]:
        pass

    @staticmethod
    def type_attributes() -> list[str]:
        return ["type_uuid", "type_id"]

    @staticmethod
    @abstractmethod
    def get_enum() -> EntitiesEnum:
        pass


class SpTypeMixin(HasTypeMixin):
    @property
    def capex(self):
        return self.data["capex"]

    @property
    def opex(self):
        return self.data["opex"]

    @property
    def s_rated(self):
        return self.data["s_rated"]

    @property
    def cos_phi_rated(self):
        return self.data["cos_phi_rated"]

    @staticmethod
    @abstractmethod
    def entity_attributes() -> list[str]:
        pass

    @staticmethod
    def type_attributes() -> list[str]:
        return HasTypeMixin.type_attributes() + [
            "capex",
            "opex",
            "s_rated",
            "cos_phi_rated",
        ]



from functools import partial
from typing import Optional, Union

import numpy as np
import seaborn as sns
from matplotlib import pyplot as plt
from matplotlib.axes import Axes

from psdmpy.models.gwr import GridWithResults
from psdmpy.models.input.participant.participant import (
    SystemParticipantsWithCapacity,
)
from psdmpy.models.result.container.participants import (
    ParticipantsResultContainer,
)
from psdmpy.models.result.grid.enhanced_node import EnhancedNodesResult
from psdmpy.models.result.participant.pq_dict import (
    PQResultDict,
    PQWithSocResultDict,
)
from psdmpy.models.result.power import PQResult, PQWithSocResult
from psdmpy.plots.common.line_plot import ax_plot_time_series
from psdmpy.plots.common.utils import (
    BLUE,
    COLOR_PALETTE,
    FIGSIZE,
    FILL_ALPHA,
    GREEN,
    LOAD_COLOR,
    ORANGE,
    PV_COLOR,
    RED,
    add_to_kwargs_if_not_exist,
    get_label_and_color,
    get_label_and_color_dict,
    legend_with_distinct_labels,
    plot_resample,
    set_date_format_and_label,
    set_subplot_title,
    set_title,
    set_xlabels_rotated,
    set_ylabel,
)


def plot_apparent_power_components(
    pq: PQResult, resolution: str, title: str | None = None
):
    """
    Plots apparent power components (magnitude, active, reactive, angle) for a given
    pq result.

    Args
        pq: PQResult
        resolution: Resolution of the plot
        title: Optional title of the plot
    """
    fig, axs = plt.subplots(4, 1, figsize=(15, 16))

    fig.subplots_adjust(hspace=0.4)

    ax_plot_power_mag(axs[0], pq, resolution=resolution, color=RED)
    axs[0].set_title("Apparent Power")
    ax_plot_active_power(axs[1], pq, resolution=resolution, color=GREEN)
    axs[1].set_title("Active Power")
    ax_plot_reactive_power(axs[2], pq, resolution=resolution, color=BLUE)
    axs[2].set_title("Reactive Power")
    ax_plot_power_angle(axs[3], pq, resolution=resolution, color=ORANGE)
    axs[3].set_title("Power Angle")

    if title is None:
        name = pq.name if pq.name else pq.input_model
        title = "Apparent Power Composition: " + name

    fig.suptitle(title, fontsize=16, y=0.93)

    return fig, axs


def plot_all_nodal_ps_branch_violin(
    gwr: GridWithResults,
):
    """
    Plots active power violin plots for all nodes across all branches.

    Args:
        gwr: GridWithResults object

    Returns:
        fig, axes
    """
    branches = gwr.grid.raw_grid.get_branches()
    nodes_res = gwr.build_enhanced_nodes_result()
    width, height = FIGSIZE
    height = height * len(branches)
    fig, axes = plt.subplots(nrows=len(branches), figsize=(width, height))
    for i, branch in enumerate(branches):
        ax_plot_nodal_ps_violin(axes[i], nodes_res, branch)
        set_subplot_title(axes[i], f"Nodal Actice Power Along Branch {i+1}")
    plt.tight_layout()

    return fig, axes


def plot_nodal_ps_violin(
    enhanced_nodes_res: EnhancedNodesResult,
    nodes: Optional[list[str]],
):
    """
    Plots violin plots for all given nodes.

    Args:
        enhanced_nodes_res: EnhancedNodesResult to plot.
        nodes: Optional list of node uuids that should be plotted. Order is preserved.

    Returns:
        fig, ax
    """
    fig, ax = plt.subplots(figsize=FIGSIZE)
    ax_plot_nodal_ps_violin(ax, enhanced_nodes_res, nodes)
    set_title(ax, "Nodal Active Power")
    return fig, ax


def ax_plot_nodal_ps_violin(
    ax: Axes,
    nodes_res: EnhancedNodesResult,
    nodes: Optional[list[str]],  # branches can be found by GridContainer.get_branches()
):
    """
    Plots violin plots for given nodes. If no nodes are passed all nodes are plotted.

    Args:
        ax: Axes object
        nodes_res: NodesResult or EnhancedNodesResult object
        nodes: Optional list of node uuids that should be plotted. Order is preserved.
    """

    if nodes:
        # get v_mag in listed sequence
        p = nodes_res.subset(nodes).p.reindex(columns=nodes)
    else:
        p = nodes_res.p

    sns.violinplot(p, showmedians=True, ax=ax, linewidth=0.5, palette=COLOR_PALETTE)

    # set labels
    uuid_to_id = nodes_res.uuid_to_id_map()
    x_labels = p.columns.map(lambda uuid: uuid_to_id[uuid])
    set_xlabels_rotated(ax, x_labels, ha="right")
    set_ylabel(ax, "Nodal active power in MW")
    _ = ax.set_xticklabels(x_labels, rotation=45, ha="right")


def plot_comparison(
    res_a: PQResult,
    res_b: PQResult,
    label_a: str,
    label_b: str,
    title: str,
    resolution: str,
    hourly_mean=False,
    flex_signal: PQResult = None,
):
    subplot_count = 2 if flex_signal is None else 3
    fig, axs = plt.subplots(
        subplot_count, 1, figsize=FIGSIZE, sharex=True, sharey=False
    )
    ax_plot_active_power(
        axs[0], res_a, resolution, hourly_mean, label=label_a, color=LOAD_COLOR
    )
    ax_plot_active_power(
        axs[0], res_b, resolution, hourly_mean, label=label_b, color=PV_COLOR
    )
    residual = res_a - res_b
    ax_plot_active_power(
        axs[1], residual, resolution, hourly_mean, label="Residual Load"
    )
    if flex_signal:
        ax_plot_active_power(
            axs[2], flex_signal, resolution, hourly_mean, label="Flex Signal"
        )
    fig.suptitle(title)
    [ax.legend() for ax in axs]
    return axs, fig


def plot_em(
    em_participant_results: ParticipantsResultContainer,
    resolution: str,
    hourly_mean: bool = False,
):
    em_uuid = list(em_participant_results.ems.entities.keys())[0]

    title = f"Household: {em_uuid}"

    fig, axs = plt.subplots(2, 1, figsize=(10, 5), sharex=True, sharey=True)
    axs[0].set_title(title)

    ax_plot_participants(
        axs[0], em_participant_results, resolution, hourly_mean=hourly_mean, stack=True
    )
    ax_plot_active_power(
        axs[1],
        em_participant_results.ems.sum(),
        resolution,
        hourly_mean=hourly_mean,
        fill_from_index=True,
    )
    return fig


def plot_aggregated_load_and_generation(
    participants: Union[ParticipantsResultContainer, list[PQResult]],
    title: str,
    resolution: str,
    hourly_mean: bool,
    with_residual=True,
):
    if with_residual:
        fig, axs = plt.subplots(2, 1, figsize=(10, 5), sharex=True, sharey=True)
        pq_results = _get_pq_results_from_union(participants)
        all_load, all_generation = zip(
            *[pq_result.divide_load_generation() for pq_result in pq_results]
        )
        agg_load = PQResult.sum(all_load)
        agg_generation = PQResult.sum(all_generation)
        ax_plot_active_power(
            axs[0],
            agg_load,
            resolution,
            hourly_mean,
            fill_from_index=True,
            color=LOAD_COLOR,
            label="Aggregated Load",
        )
        ax_plot_active_power(
            axs[0],
            agg_generation,
            resolution,
            hourly_mean,
            fill_from_index=True,
            color=PV_COLOR,
            label="Aggregated Generation",
        )
        ax_plot_active_power(
            axs[1],
            PQResult.sum([agg_load, agg_generation]),
            resolution,
            hourly_mean,
            fill_from_index=True,
            color=LOAD_COLOR,
            label="Residual Load",
        )
        set_title(axs[0], title)
        axs[0].legend(loc=7)
        axs[1].legend()
        return fig, axs


def plot_all_participants(
    participants: Union[ParticipantsResultContainer, list[PQResult]],
    title: str,
    resolution: str,
    hourly_mean: bool,
    stack=False,
    with_residual=False,
    **kwargs,
):
    if with_residual:
        fig, axs = plt.subplots(2, 1, figsize=(10, 5), sharex=True, sharey=True)
        ax_plot_participants(
            axs[0], participants, resolution, hourly_mean, stack, **kwargs
        )
        participants_sum = (
            participants.sum()
            # here I would usually check if participants is a ParticipantsResultContainer
            # but for some weird reason this returns False every time
            if not isinstance(participants, list)
            else PQResult.sum(participants)
        )
        ax_plot_active_power(
            axs[1],
            participants_sum,
            resolution,
            hourly_mean,
            fill_from_index=True,
            **kwargs,
        )
        set_title(axs[0], title)
    else:
        fig, axs = plt.subplots(figsize=FIGSIZE)
        ax_plot_participants(
            axs, participants, resolution, hourly_mean, stack, **kwargs
        )
        set_title(axs, title)
    return fig, axs


def _get_pq_results_from_union(
    participants: Union[ParticipantsResultContainer, list[PQResult]]
) -> list[PQResult]:
    return (
        participants
        if isinstance(participants, list)
        else [
            participant.sum()
            for participant in participants.to_list(
                include_em=False, include_flex=False
            )
            if participant
        ]
    )


def ax_plot_participants(
    ax: Axes,
    participants: Union[ParticipantsResultContainer, list[PQResult]],
    resolution: str,
    hourly_mean: bool = False,
    stack=False,
    **kwargs,
):
    pq_results = _get_pq_results_from_union(participants)
    if stack:
        plot_kwargs = [kwargs for _ in range(len(pq_results))]
        ax_plot_stacked_pq(ax, pq_results, resolution, hourly_mean, plot_kwargs)
    else:
        [
            ax_plot_active_power(ax, pq_result, resolution, hourly_mean, **kwargs)
            for pq_result in pq_results
        ]
    legend_with_distinct_labels(ax)


def ax_plot_stacked_pq(
    ax: Axes,
    results: list[PQResult],
    resolution: str,
    hourly_mean: bool = False,
    plot_kwargs: list[dict] = None,
):
    residual_load, residual_generation = results[0].divide_load_generation()

    plot_partial = partial(
        ax_plot_active_power,
        ax=ax,
        resolution=resolution,
        hourly_mean=hourly_mean,
    )

    plot_partial(
        res=residual_load, fill_from_index=True, **plot_kwargs[0] if plot_kwargs else {}
    )
    plot_partial(
        res=residual_generation,
        fill_from_index=True,
        **plot_kwargs[0] if plot_kwargs else {},
    )

    for idx, res in enumerate(results[1:]):
        load, generation = res.divide_load_generation()

        if load.p.sum() > 0:
            load_sum = PQResult(
                res.entity_type, "", "", PQResult.sum([load, residual_load]).data
            )
            plot_partial(
                res=load_sum,
                fill_between=residual_load.p,
                **plot_kwargs[idx + 1] if plot_kwargs else {},
            )
            residual_load = load_sum

        if generation.p.sum() < 0:
            generation_sum = PQResult(
                res.entity_type,
                "",
                "",
                PQResult.sum([generation, residual_generation]).data,
            )
            plot_partial(
                res=generation_sum,
                fill_between=residual_generation.p,
                **plot_kwargs[idx + 1] if plot_kwargs else {},
            )
            residual_generation = generation_sum
    handles, labels = ax.get_legend_handles_labels()
    by_label = dict(zip(labels, handles))
    ax.legend(by_label.values(), by_label.keys())


def plot_participants_sum(
    res: PQResultDict,
    title: str,
    resolution: str,
    hourly_mean: bool = False,
    fill_from_index: bool = True,
    **kwargs,
):
    if not isinstance(res, PQResultDict):
        raise TypeError(
            "Data must be of type ParticipantsResult but is {}".format(type(res))
        )
    fig, ax = plt.subplots(figsize=FIGSIZE)
    ax_plot_active_power(
        ax, res.sum(), resolution, hourly_mean, fill_from_index, **kwargs
    )
    set_title(ax, title)
    return fig


def plot_participants_with_soc_sum(
    res: PQWithSocResultDict,
    input: SystemParticipantsWithCapacity,
    title: str,
    resolution: str,
    hourly_mean: bool = False,
    fill_from_index: bool = True,
    **kwargs,
):
    fig, axs = plt.subplots(2, 1, figsize=FIGSIZE, sharex=True, sharey=False)
    axs[0].set_title(title)
    sum = PQWithSocResultDict.sum_with_soc(res, input)
    ax_plot_active_power(
        axs[0],
        sum,
        resolution,
        hourly_mean=hourly_mean,
        fill_from_index=fill_from_index,
        **kwargs,
    )
    ax_plot_soc(axs[1], sum, resolution, hourly_mean, fill_from_index, **kwargs)
    return fig


def plot_active_power_with_soc(
    res: PQWithSocResult,
    title: str,
    resolution: str,
    hourly_mean=False,
    fill_from_index=False,
    **kwargs,
):
    fig, axs = plt.subplots(2, 1, figsize=FIGSIZE, sharex=True, sharey=False)
    axs[0].set_title(title)
    ax_plot_active_power(
        axs[0],
        res,
        resolution,
        hourly_mean=hourly_mean,
        fill_from_index=fill_from_index,
        set_x_label=False,
        **kwargs,
    )
    ax_plot_soc(axs[1], res, resolution, hourly_mean, fill_from_index, **kwargs)
    return fig


def plot_active_power(
    res: PQResult,
    resolution: str,
    title: Optional[str] = None,
    hourly_mean: bool = False,
    fill_from_index=True,
    **kwargs,
):
    fig, ax = plt.subplots(figsize=FIGSIZE)
    ax.set_ylabel("Power in MW")
    ax_plot_active_power(
        ax, res, resolution, hourly_mean, fill_from_index=fill_from_index, **kwargs
    )
    if not title:
        title = "Active power of {}".format(res.entity_type.get_plot_name())
    set_title(ax, title)
    return fig, ax


def ax_plot_active_power(
    ax: Axes,
    res: PQResult,
    resolution: str,
    hourly_mean: bool = False,
    fill_from_index: bool = False,
    fill_between=None,
    set_x_label=True,
    **kwargs,
):
    if len(res.p) == 0:
        raise ValueError("Active power time series is empty. No data to plot")

    ax = ax_plot_time_series(
        ax,
        res.p,
        res.entity_type,
        resolution,
        hourly_mean=hourly_mean,
        fill_from_index=fill_from_index,
        fill_between=fill_between,
        set_x_label=set_x_label,
        **kwargs,
    )
    ax.set_ylabel("Power in MW")


def ax_plot_reactive_power(
    ax: Axes,
    res: PQResult,
    resolution: str,
    hourly_mean: bool = False,
    fill_from_index: bool = False,
    fill_between=None,
    set_x_label=True,
    **kwargs,
):
    if len(res.q) == 0:
        raise ValueError("Reactive power time series is empty. No data to plot")

    ax = ax_plot_time_series(
        ax,
        res.q,
        res.entity_type,
        resolution,
        hourly_mean=hourly_mean,
        fill_from_index=fill_from_index,
        fill_between=fill_between,
        set_x_label=set_x_label,
        **kwargs,
    )
    ax.set_ylabel("Reactive Power in MVar")


def ax_plot_power_angle(
    ax: Axes,
    res: PQResult,
    resolution: str,
    hourly_mean: bool = False,
    fill_from_index: bool = False,
    fill_between=None,
    set_x_label=True,
    **kwargs,
):
    angle = res.angle()
    if len(angle) == 0:
        raise ValueError("Angle time series is empty. No data to plot")

    ax = ax_plot_time_series(
        ax,
        angle,
        res.entity_type,
        resolution,
        hourly_mean=hourly_mean,
        fill_from_index=fill_from_index,
        fill_between=fill_between,
        set_x_label=set_x_label,
        **kwargs,
    )
    ax.set_ylabel("Angle in degree")


def ax_plot_power_mag(
    ax: Axes,
    res: PQResult,
    resolution: str,
    hourly_mean: bool = False,
    fill_from_index: bool = False,
    fill_between=None,
    set_x_label=True,
    **kwargs,
):
    complex_mag = res.complex_power().apply(lambda x: np.abs(x))
    if len(complex_mag) == 0:
        raise ValueError("Angle time series is empty. No data to plot")

    ax = ax_plot_time_series(
        ax,
        complex_mag,
        res.entity_type,
        resolution,
        hourly_mean=hourly_mean,
        fill_from_index=fill_from_index,
        fill_between=fill_between,
        set_x_label=set_x_label,
        **kwargs,
    )
    ax.set_ylabel("Apparent Power Magnitude in MVA")


def ax_plot_soc(
    ax: Axes,
    res: PQWithSocResult,
    resolution: str,
    hourly_mean: bool = False,
    fill_from_index=False,
    **kwargs,
):
    label, color = get_label_and_color(res.entity_type)
    set_date_format_and_label(ax, resolution)
    ax.set_ylabel("SOC in percent")
    soc = plot_resample(res.soc, hourly_mean)
    ax.plot(soc, color=color, label=label, **kwargs)
    ax.set_ylim(bottom=0, top=100)
    if fill_from_index:
        ax.fill_between(soc.index, soc, alpha=FILL_ALPHA, color=color)


def plot_sorted_annual_load_duration(
    res: PQResult, s_rated_mw: float = None, fill_from_index=False, **kwargs
):
    args = get_label_and_color_dict(res.entity_type)
    kwargs = add_to_kwargs_if_not_exist(kwargs, args)
    fig, ax = plt.subplots(figsize=FIGSIZE)
    annual_duartion_series = res.annual_duration_series()
    plt.plot(annual_duartion_series, **kwargs)
    set_title(ax, f"{kwargs['label']} Annual Load Duration Curve")
    ax.set_xlabel("Duration in hours")
    ax.set_ylabel("Rated power")
    if fill_from_index:
        ax.fill_between(
            annual_duartion_series.index,
            annual_duartion_series,
            alpha=FILL_ALPHA,
            color=kwargs["color"],
        )
    if s_rated_mw:
        # todo: implement right hand y-axis for displaying s_rated percentage
        pass
    return fig, ax



import copy
import logging
from abc import ABC, abstractmethod
from dataclasses import dataclass, replace
from datetime import datetime
from typing import Optional, Tuple, TypeVar, Union

import pandas as pd
from pandas import DataFrame, Series

from psdmpy.errors import ComparisonError
from psdmpy.io.utils import to_date_time
from psdmpy.models.enums import EntitiesEnum, SystemParticipantsEnum
from psdmpy.models.input.entity import EntityType
from psdmpy.processing.dataframe import (
    compare_dfs,
    filter_data_for_time_interval,
)

ResultType = TypeVar("ResultType", bound="ResultEntities")


@dataclass(frozen=True)
class ResultEntities(ABC):
    """
    Abstract base class for all result entities. Results are time series data, which can be mapped to their respective input entities via `input_model`.
    The time series is event discrete. Which means every state is valid until the next state is reached. The time series data is stored in the `data` attribute.
    Attributes:
        entity_type: The type of the entity.
        input_model: The input model uuid the results belong to.
        name: The id of the entity. Can be None.
        data: Resulting time series data of the entity.
    """

    entity_type: EntitiesEnum
    input_model: str
    name: Optional[str]
    data: DataFrame

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False
        if self.entity_type != other.entity_type:
            return False
        if self.input_model != other.input_model:
            return False
        if self.name != other.name:
            return False
        try:
            compare_dfs(self.data, other.data)
        except ComparisonError:
            return False

    def __repr__(self):
        return self.data

    def __len__(self):
        return len(self.data)

    def __getitem__(self, where: Union[slice, datetime, list[datetime]]):
        if isinstance(where, slice):
            start, stop, step = where.start, where.stop, where.step
            if step is not None:
                logging.warning("Step is not supported for slicing. Ignoring it.")
            if not (isinstance(start, datetime) and isinstance(stop, datetime)):
                raise ValueError("Only datetime slicing is supported")
            return self.filter_for_time_interval(start, stop)
        elif isinstance(where, datetime):
            filtered, dt = self._get_data_by_datetime(where)
            data = pd.DataFrame(filtered).T
            return self.build(self.entity_type, self.input_model, data, dt, self.name)
        elif isinstance(where, list):
            filtered = [self._get_data_by_datetime(time)[0] for time in where]
            data = pd.DataFrame(pd.concat(filtered, axis=1)).T
            max_dt = sorted(data.index)[-1]
            return self.build(
                self.entity_type, self.input_model, data, max_dt, name=self.name
            )

    def _get_data_by_datetime(self, dt: datetime) -> Tuple[Series, datetime]:
        if dt > self.data.index[-1]:
            logging.warning(
                "Trying to access data after last time step. Returning last time step."
            )
            return self.data.iloc[-1], self.data.index[-1]
        if dt < self.data.index[0]:
            logging.warning(
                "Trying to access data before first time step. Returning first time step."
            )
            return self.data.iloc[0], self.data.index[0]
        else:
            return self.data.asof(dt), dt

    def compare(self, other) -> None:
        """
        Compares the current Entities instance with another Entities instance.

        Args:
            other: The other Entities instance to compare with.

        Returns:
            None
        """
        if not isinstance(other, type(self)):
            raise ComparisonError(
                f"Type of self: {type(self)} != type of other: {type(other)}", errors=[]
            )

        differences = []

        if self.entity_type != other.entity_type:
            differences.append(
                f"Entity type left: {self.entity_type} != right {other.entity_type}"
            )
        if self.input_model != other.input_model:
            differences.append(
                f"Input model left: {self.input_model} != right {other.input_model}"
            )
        if self.name != other.name:
            differences.append(f"Name left: {self.name} != right {other.name}")

        try:
            compare_dfs(self.data, other.data)
        except AssertionError as e:
            raise ComparisonError(
                f"{self.entity_type.get_plot_name()} entities are not equal.",
                errors=[(type(self), str(e))],
            )

    def copy(
        self: ResultType,
        deep=True,
        **changes,
    ) -> ResultType:
        """
        Creates a copy of the current ResultEntities instance.
        By default does a deep copy of all data and replaces the given changes.
        When deep is false, only the references to the data of the non-changed attribtues are copied.

        Args:
            deep: Whether to do a deep copy of the data.
            **changes: The changes to apply to the copy.

        Returns:
            The copy of the current Entities instance.
        """
        to_copy = copy.deepcopy(self) if deep else self
        return replace(to_copy, **changes)

    @staticmethod
    @abstractmethod
    def attributes() -> list[str]:
        """
        Returns a list of all attributes of the corresponding PSDM entity. The attributes correspond to its data columns.
        """
        pass

    @classmethod
    def create_empty(
        cls, entity_type: EntitiesEnum, input_model: str, name: Optional[str] = None
    ) -> ResultType:
        """
        Creates an empty ResultEntities object with the given entity type and optionally name.
        Args:
            entity_type: The entity type of the ResultEntities object.
            input_model: The input model of the ResultEntities object.
            name: The name or id of the corresponding input object. Can be none.
        Returns:
            An empty ResultEntities object.
        """
        data = cls.empty_data()
        return cls(entity_type, input_model, name, data)

    @classmethod
    def empty_data(cls, index=None):
        """
        Returns an empty DataFrame with the attributes as columns.
        """
        return (
            pd.DataFrame(columns=cls.attributes(), index=index)
            if index
            else pd.DataFrame(columns=cls.attributes())
        )

    @classmethod
    def build(
        cls,
        entity_type: EntitiesEnum,
        input_model: str,
        data: DataFrame,
        end: datetime,
        name: Optional[str] = None,
    ) -> "ResultEntities":
        """
        Creates a ResultEntities object from the given data.
        The end time is used to fill the last time step of the data with the end time.
        This is because we have a time discrete simulation, which means the last entry in the data
        is the state of the entity at the end of the simulation.
        Args:
            entity_type: The entity type of the ResultEntities object.
            input_model: The input model for which the result data was calculated.
            data: The data of the ResultEntities object.
            end: The end time of the simulation.
            name: The name or id of the corresponding input object. Can be none.
        """
        if data.empty:
            return cls.create_empty(entity_type, input_model, name)

        if "time" in data.columns:
            data["time"] = data["time"].apply(
                lambda date_string: to_date_time(date_string)
            )
            data = data.set_index("time", drop=True)

        last_state = data.iloc[len(data) - 1]
        if last_state.name != end:
            last_state.name = end
            data = pd.concat([data, DataFrame(last_state).transpose()])
        # todo: deal with duplicate indexes -> take later one
        data = data[~data.index.duplicated(keep="last")]
        data.sort_index(inplace=True)
        return cls(entity_type, input_model, name, data)

    # TODO: Check if end time is in or excluded
    def filter_for_time_interval(
        self, start: datetime, end: datetime
    ) -> "ResultEntities":
        """
        Filters the data for the given time interval. The data can also be filtered via object[datetime:datetime].
        See __getitem__ for more information.
        Args:
            start: The start time of the time interval.
            end: The end time of the time interval.
        Returns:
            A new ResultEntities object with the filtered data.
        """
        filtered_data = filter_data_for_time_interval(self.data, start, end)
        return self.build(
            self.entity_type, self.input_model, filtered_data, end, self.name
        )

    @classmethod
    # todo: find a way for parallel calculation
    def sum(cls, results: list[ResultType]) -> ResultType:
        """
        Sums up the time series data for a list of ResultEntities objects.
        Args:
            results: A list of ResultEntities objects.
        Returns:
            A new ResultEntities object with the summed up data.
        """
        if len(results) == 0:
            return cls.create_empty(SystemParticipantsEnum.PARTICIPANTS_SUM, "", "")
        if len(results) == 1:
            return results[0]
        agg = results[0]
        for result in results[1::]:
            agg += result
        return agg

    def find_input_entity(self, input_model: EntityType) -> EntityType:
        """
        Finds the input entity for the object.
        Args:
            input_model: The input models within which the input entity should be found.
        Returns:
            The filtered input entities.
        """
        if self.entity_type != input_model.get_enum():
            logging.warning("Input model type does not match result type!")
        return input_model.subset([self.input_model])



import os.path
from dataclasses import dataclass
from datetime import datetime
from typing import Dict

import pandas as pd
from pandas import DataFrame

from psdmpy.models.enums import SystemParticipantsEnum
from psdmpy.models.result.flex_option import FlexOptionResult
from psdmpy.models.result.participant.dict import ResultDict


@dataclass(frozen=True)
class FlexOptionsResult(ResultDict):
    entities: Dict[str, FlexOptionResult]

    def to_csv(self, output_path: str):
        self.to_df().to_csv(
            os.path.join(output_path, "flex_df.csv"), index_label="time"
        )

    def to_df(self) -> DataFrame:
        return pd.concat(
            [f.data for f in self.entities.values()],
            keys=self.entities.keys(),
            axis=1,
        ).ffill()

    def to_multi_index_df(
        self, participants_res  # type hinting leads to circular import
    ) -> DataFrame:
        flex_midfs = {}
        for res in participants_res.to_list():
            uuids = res.participants.keys()
            flex_res = self.subset(uuids)
            if flex_res:
                flex_dfs = []
                participant_uuids = []
                [
                    (participant_uuids.append(uuid), flex_dfs.append(flex.data))
                    for uuid, flex in flex_res.entities.items()
                ]
                flex_midf = pd.concat(flex_dfs, keys=participant_uuids, axis=1)
                flex_midfs[res.entity_type.value] = flex_midf
        return pd.concat(flex_midfs.values(), keys=(flex_midfs.keys()), axis=1).ffill()

    def sum(self) -> FlexOptionResult:
        return FlexOptionResult.sum(list(self.entities.values()))

    @classmethod
    def from_csv(
        cls,
        sp_type: SystemParticipantsEnum,
        simulation_data_path: str,
        delimiter: str,
        simulation_end: datetime,
        from_df: bool = False,
    ) -> "FlexOptionsResult":
        if from_df:
            path = os.path.join(simulation_data_path, "flex_df.csv")
            if not os.path.exists(path):
                return cls.create_empty(SystemParticipantsEnum.FLEX_OPTIONS)
            data = pd.read_csv(
                os.path.join(simulation_data_path, "flex_df.csv"),
                delimiter=delimiter,
                index_col=0,
                header=[0, 1],
            )
            data_dicts = {}
            for uuid in data.columns.get_level_values(0):
                data_dicts[uuid] = data[uuid]
            return cls(SystemParticipantsEnum.FLEX_OPTIONS, data_dicts)

        participant_grpd_df = ResultDict.get_grpd_df(
            sp_type,
            simulation_data_path,
            delimiter,
        )

        if not participant_grpd_df:
            return cls.create_empty(SystemParticipantsEnum.FLEX_OPTIONS)

        participants = dict(
            participant_grpd_df.apply(
                lambda grp: FlexOptionResult.build(
                    sp_type,
                    grp.name,
                    grp.drop(columns=["input_model"]),
                    simulation_end,
                )
            )
        )
        return cls(sp_type, participants)



import logging
from typing import Optional, Union

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.axes import Axes

from psdmpy.models.gwr import GridWithResults
from psdmpy.models.input.connector.lines import Lines
from psdmpy.models.result.grid.connector import ConnectorResult
from psdmpy.models.result.grid.transformer import Transformer2WResult
from psdmpy.plots.common.line_plot import ax_plot_time_series
from psdmpy.plots.common.utils import (
    BLUE,
    FIGSIZE,
    ax_plot_secondary_axis,
    set_subplot_title,
    set_suptitle,
    set_title,
)


def plot_transformer_s(
    res: Transformer2WResult,
    side: str,
    gwr: GridWithResults,
    resolution: str,
    include_utilisation: bool = True,
    show_util_grid_lines: bool = False,
    title: Optional[str] = None,
    fill_from_index: bool = False,
    fill_between=None,
    set_x_label=True,
    **kwargs,
):
    transformer_name = res.name if res.name else res.input_model
    if title is None:
        title = f"Rated Power Transformer: {transformer_name}"
    if include_utilisation:
        width, height = FIGSIZE
        fig, axs = plt.subplots(2, 1, figsize=(width, height * 2), sharex=False)
        ax_plot_transformer_s(
            axs[0],
            res,
            side,
            gwr,
            resolution,
            fill_from_index=fill_from_index,
            fill_between=fill_between,
            set_x_label=set_x_label,
            **kwargs,
        )
        set_subplot_title(axs[0], "Rated Power Magnitude")
        ax_plot_transformer_utilization(
            axs[1],
            res,
            side,
            gwr,
            resolution,
            fill_from_index=fill_from_index,
            fill_between=fill_between,
            set_x_label=False,
            color=BLUE,
            **kwargs,
        )
        set_subplot_title(axs[1], "Transformer Utilization")
        fig.suptitle(title, fontsize=16)
        fig.subplots_adjust(hspace=0.4)
        return fig, axs
    else:
        fig, ax = plt.subplots(figsize=FIGSIZE)
        ax_plot_transformer_s(
            ax,
            res,
            side,
            gwr,
            resolution,
            fill_from_index=fill_from_index,
            fill_between=fill_between,
            set_x_label=set_x_label,
            **kwargs,
        )
        set_suptitle(fig, title)
        return fig, ax


def ax_plot_transformer_s(
    ax: Axes,
    res: Transformer2WResult,
    side: str,
    gwr: GridWithResults,
    resolution: str,
    fill_from_index: bool = False,
    fill_between=None,
    set_x_label=True,
    **kwargs,
):
    if len(res.i_a_mag) == 0:
        raise ValueError("Transformer current time series is empty. No data to plot")

    rated_power = res.calc_apparent_power_gwr(gwr, side).apply(lambda x: np.real(x))
    ax_plot_time_series(
        ax,
        rated_power,
        res.entity_type,
        resolution,
        fill_from_index=fill_from_index,
        fill_between=fill_between,
        set_x_label=set_x_label,
        **kwargs,
    )
    ax.set_ylabel("Rated Power in kVA")


def ax_plot_transformer_utilization(
    ax: Axes,
    res: Transformer2WResult,
    side: str,
    gwr: GridWithResults,
    resolution: str,
    fill_from_index: bool = False,
    fill_between=None,
    set_x_label=True,
    **kwargs,
):
    ax_plot_time_series(
        ax,
        res.calc_transformer_utilisation(gwr, side),
        res.entity_type,
        resolution,
        fill_from_index=fill_from_index,
        fill_between=fill_between,
        set_x_label=set_x_label,
        **kwargs,
    )
    ax.set_ylabel("Transformer Utilization")


def plot_line_current(
    res: ConnectorResult,
    side: str,
    resolution: str,
    include_utilisation: bool = False,
    i_max_src: Optional[Union[float, Lines]] = None,
    title: Optional[str] = None,
    fill_from_index: bool = False,
    fill_between=None,
    set_x_label=True,
    **kwargs,
):
    line_name = res.name if res.name else res.input_model
    if title is None:
        title = f"Current Magnitude Line: {line_name}"
    fig, ax = plt.subplots(figsize=FIGSIZE)
    ax_plot_line_current(
        ax,
        res,
        side,
        resolution,
        i_max_src=i_max_src,
        include_utilisation=include_utilisation,
        fill_from_index=fill_from_index,
        fill_between=fill_between,
        set_x_label=set_x_label,
        **kwargs,
    )
    ax.set_ylabel("Current in Ampere")
    set_title(ax, title)
    return fig, ax


def ax_plot_line_current(
    ax: Axes,
    res: ConnectorResult,
    side: str,
    resolution: str,
    include_utilisation: bool = True,
    i_max_src: Optional[Union[float, Lines]] = None,
    fill_from_index: bool = False,
    fill_between=None,
    set_x_label=True,
    **kwargs,
):
    if len(res.i_a_mag) == 0:
        raise ValueError("Line current time series is empty. No data to plot")

    current = get_connector_current(side, res)

    ax = ax_plot_time_series(
        ax,
        current,
        res.entity_type,
        resolution,
        fill_from_index=fill_from_index,
        fill_between=fill_between,
        set_x_label=set_x_label,
        **kwargs,
    )
    ax.set_ylabel("Current Magnitude in Ampere")

    if include_utilisation:
        if i_max_src is None:
            logging.warning("No i_max_src provided. Cannot plot utilisation")
            return
        line_i_max = (
            i_max_src
            if isinstance(i_max_src, float)
            else i_max_src.subset(res.input_model).i_max.iloc[0]
        )
        utilisation = current / line_i_max
        ax_plot_secondary_axis(
            ax, utilisation, "Line Utilisation", show_secondary_grid_lines=True
        )


def plot_connector_current(
    res: ConnectorResult,
    side: str,
    resolution: str,
    title: Optional[str] = None,
    fill_from_index: bool = False,
    fill_between=None,
    set_x_label=True,
    **kwargs,
):
    connector_type = res.entity_type.get_plot_name()
    connector_name = res.name if res.name else res.input_model
    if title is None:
        title = f"Current Magnitude {connector_type}: {connector_name}"
    fig, ax = plt.subplots(figsize=FIGSIZE)
    ax_plot_connector_current(
        ax,
        res,
        side,
        resolution,
        fill_from_index=fill_from_index,
        fill_between=fill_between,
        set_x_label=set_x_label,
        **kwargs,
    )
    ax.set_ylabel("Current in Ampere")
    set_title(ax, title)
    return fig, ax


def ax_plot_connector_current(
    ax: Axes,
    res: ConnectorResult,
    side: str,
    resolution: str,
    fill_from_index: bool = False,
    fill_between=None,
    set_x_label=True,
    **kwargs,
):
    connector_type = res.entity_type.get_plot_name()
    if len(res.i_a_mag) == 0:
        raise ValueError(
            f"{connector_type} current time series is empty. No data to plot"
        )

    current = get_connector_current(side, res)

    ax = ax_plot_time_series(
        ax,
        current,
        res.entity_type,
        resolution,
        fill_from_index=fill_from_index,
        fill_between=fill_between,
        set_x_label=set_x_label,
        **kwargs,
    )
    ax.set_ylabel("Current Magnitude in Ampere")


def ax_plot_connector_angle(
    ax: Axes,
    res: ConnectorResult,
    side: str,
    resolution: str,
    fill_from_index: bool = False,
    fill_between=None,
    set_x_label=True,
    **kwargs,
):
    connector_type = res.entity_type.get_plot_name()
    if len(res.i_a_mag) == 0:
        raise ValueError(
            f"{connector_type} angle time series is empty. No data to plot"
        )

    angle = get_connector_angle(side, res)

    ax = ax_plot_time_series(
        ax,
        angle,
        res.entity_type,
        resolution,
        fill_from_index=fill_from_index,
        fill_between=fill_between,
        set_x_label=set_x_label,
        **kwargs,
    )


def get_connector_current(side: str, res: ConnectorResult):
    if side == "a":
        return res.i_a_mag
    elif side == "b":
        return res.i_b_mag
    else:
        raise ValueError('Side should be either "a" for node a or "b" for node b')


def get_connector_angle(side: str, res: ConnectorResult):
    if side == "a":
        return res.i_a_ang
    elif side == "b":
        return res.i_b_ang
    else:
        raise ValueError('Side should be either "a" for node a or "b" for node b')



from abc import ABC
from dataclasses import dataclass
from typing import TypeVar, Union

import pandas as pd

from psdmpy.models.input.entity import Entities
from psdmpy.models.input.node import Nodes

ConnectorType = TypeVar("ConnectorType", bound="Connector")


@dataclass(frozen=True)
class Connector(Entities, ABC):
    @property
    def node_a(self):
        return self.data["node_a"]

    @property
    def node_b(self):
        return self.data["node_b"]

    @property
    def parallel_devices(self):
        return self.data["parallel_devices"]

    @property
    def node(self):
        return pd.concat([self.node_a, self.node_b])

    def filter_by_nodes(
        self: ConnectorType, node_uuids: Union[str, list[str]], both_in_nodes=False
    ) -> ConnectorType:
        """
        Returns all connectors that are connected to the given nodes.

        Args:
            node_uuids: List of node uuids to find lines for.
            both_in_nodes: If True, both nodes of the line must be in the given list of nodes.

        Returns:
            ConnectorType: Connectors that are connected to the given nodes.
        """
        if isinstance(node_uuids, str):
            node_uuids = [node_uuids]
        if both_in_nodes:
            data = self.data[
                (self.node_a.isin(node_uuids)) & (self.node_b.isin(node_uuids))
            ]
        else:
            data = self.data[
                (self.node_a.isin(node_uuids)) | (self.node_b.isin(node_uuids))
            ]
        return type(self)(data)

    def filter_by_node_pair(
        self: ConnectorType, node_a_uuid: str, node_b_uuid: str
    ) -> ConnectorType:
        """
        Returns the connector that connects the given nodes. Order of the nodes does not matter.

        Args:
            node_a_uuid (str): UUID of the first node.
            node_b_uuid (str): UUID of the second node.

        Returns:
            Lines: Line that connects the given nodes. Will be empty if no line is found.
        """
        return self.filter_by_nodes([node_a_uuid, node_b_uuid], both_in_nodes=True)

    def insert_node_id_columns(self, nodes: Nodes) -> None:
        index_to_id = nodes.id.to_dict()

        if "node_a_id" not in self.data.columns:
            self.data.insert(self.data.columns.get_loc("node_a") + 1, "node_a_id", None)
        self.data["node_a_id"] = self.node_a.map(index_to_id)

        if "node_b_id" not in self.data.columns:
            self.data.insert(self.data.columns.get_loc("node_b") + 1, "node_b_id", None)
        self.data["node_b_id"] = self.node_b.map(index_to_id)

    @staticmethod
    def attributes() -> list[str]:
        return Entities.attributes() + ["node_a", "node_b", "parallel_devices"]



from dataclasses import dataclass
from typing import List

from psdmpy.models.enums import SystemParticipantsEnum
from psdmpy.models.input.participant.participant import SystemParticipants


@dataclass(frozen=True)
class Loads(SystemParticipants):
    @staticmethod
    def get_enum() -> SystemParticipantsEnum:
        return SystemParticipantsEnum.LOAD

    @property
    def s_rated(self):
        return self.data["s_rated"]

    @property
    def standard_load_profile(self):
        return self.data["load_profile"]

    @property
    def dsm(self):
        return self.data["dsm"]

    @property
    def e_cons_annual(self):
        return self.data["e_cons_annual"]

    @property
    def cos_phi_rated(self):
        return self.data["cos_phi_rated"]

    @staticmethod
    def attributes() -> List[str]:
        return SystemParticipants.attributes() + [
            "load_profile",
            "dsm",
            "e_cons_annual",
            "s_rated",
            "cos_phi_rated",
        ]



from datetime import datetime, timedelta

import pandas as pd

from psdmpy.models.enums import (
    EntitiesEnum,
    RawGridElementsEnum,
    SystemParticipantsEnum,
)
from psdmpy.models.result.grid.node import NodeResult, NodesResult
from psdmpy.models.result.participant.pq_dict import PQResultDict
from psdmpy.models.result.power import PQResult

# These are just some early quite specific conversions. Additional ones will be added as needed.


def pp_load_to_participants_result(
    pp_net_file: str,  # pp net xlsx export
    participant_p_file: str,  # pp time series p_mw export
    participant_q_file: str,  # pp time series q_mvar export
    psdm_load_file: str,  # psdm load csv input
    start: datetime,  # start of the time series simulation
    resolution: timedelta,  # resolution of the time series simulation
) -> PQResultDict:
    return _pp_to_psdm_result(
        entity_type=SystemParticipantsEnum.LOAD,
        pp_net_file=pp_net_file,
        pp_sheet_name="load",
        pp_attribute_a_file=participant_p_file,
        psdm_attribute_a_name="p",
        pp_attribute_b_file=participant_q_file,
        psdm_attribute_b_name="q",
        psdm_entity_input_file=psdm_load_file,
        start=start,
        resolution=resolution,
        res_entity_class=PQResult,
        res_dict_class=PQResultDict,
    )


def pp_node_to_nodes_result(
    pp_net_file: str,  # pp net xlsx export
    node_vm_file: str,  # pp time series vm_pu export
    node_va_file: str,  # pp time series va_degree export
    psdm_node_file: str,  # psdm node csv input
    start: datetime,  # start of the time series simulation
    resolution: timedelta,  # resolution of the time series simulation
) -> PQResultDict:
    return _pp_to_psdm_result(
        entity_type=RawGridElementsEnum.NODE,
        pp_net_file=pp_net_file,
        pp_sheet_name="bus",
        pp_attribute_a_file=node_vm_file,
        psdm_attribute_a_name="v_mag",
        pp_attribute_b_file=node_va_file,
        psdm_attribute_b_name="v_ang",
        psdm_entity_input_file=psdm_node_file,
        start=start,
        resolution=resolution,
        res_entity_class=NodeResult,
        res_dict_class=NodesResult,
    )


def _pp_to_psdm_result(
    entity_type: EntitiesEnum,
    pp_net_file: str,
    pp_sheet_name: str,
    pp_attribute_a_file: str,
    psdm_attribute_a_name: str,
    pp_attribute_b_file: str,
    psdm_attribute_b_name: str,
    psdm_entity_input_file: str,
    start: datetime,
    resolution: timedelta,
    res_entity_class,
    res_dict_class,
):
    pp_entity = pd.read_excel(pp_net_file, sheet_name=pp_sheet_name, index_col=0)
    idx_to_id = pp_entity["name"].to_dict()
    psdm_entity = pd.read_csv(psdm_entity_input_file, sep=";", index_col=0)
    id_to_uuid = {id: uuid for uuid, id in psdm_entity["id"].to_dict().items()}
    idx_to_uuid = {idx: id_to_uuid[id] for idx, id in idx_to_id.items()}
    pp_attribute_a = pd.read_json(pp_attribute_a_file)
    pp_attribute_b = pd.read_json(pp_attribute_b_file)
    pp_attribute_a["time"] = pd.date_range(
        start=start, periods=len(pp_attribute_a), freq=resolution
    )
    pp_attribute_a.set_index("time", inplace=True)
    pp_attribute_b["time"] = pd.date_range(
        start=start, periods=len(pp_attribute_a), freq=resolution
    )
    pp_attribute_b.set_index("time", inplace=True)
    participants = {}

    for ts in pp_attribute_a.columns:
        data = pd.concat(
            [
                pp_attribute_a[ts].rename(psdm_attribute_a_name),
                pp_attribute_b[ts].rename(psdm_attribute_b_name),
            ],
            axis=1,
        )
        uuid = idx_to_uuid[ts]
        node_res = res_entity_class(
            type=entity_type, name=idx_to_id[ts], input_model=uuid, data=data
        )
        participants[uuid] = node_res

    return res_dict_class(entity_type=entity_type, entities=participants)



import numpy as np
from pandas import Series


def rmse_by_iloc(predictions: Series, targets: Series):
    """
    Calculates the RMSE between two series. Resets the index to calculate by their index (iloc)

    Args:
        predictions: The predictions series.
        targets: The targets series.

    Returns:
        The RMSE between the two series.
    """
    if len(predictions) != len(targets):
        raise ValueError("Predictions and targets should have the same length!")
    return np.sqrt(
        (
            (predictions.reset_index(drop=True) - targets.reset_index(drop=True)) ** 2
        ).mean()
    )



from dataclasses import dataclass

from psdmpy.models.input.container.mixins import ContainerMixin
from psdmpy.models.input.thermal.bus import ThermalBus
from psdmpy.models.input.thermal.house import ThermalHouse


@dataclass
class ThermalGridContainer(ContainerMixin):
    busses: ThermalBus
    houses: ThermalHouse

    def to_list(self, include_empty: bool = False):
        lst = [self.busses, self.houses]
        if include_empty:
            return lst
        else:
            return [x for x in lst if x]

    @classmethod
    def from_csv(cls, path: str, delimiter: str):
        busses = ThermalBus.from_csv(path, delimiter)
        houses = ThermalHouse.from_csv(path, delimiter)
        return cls(busses, houses)

    @classmethod
    def create_empty(cls):
        return cls(ThermalBus.create_empty(), ThermalHouse.create_empty())



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, List

import numpy as np
import pandas as pd
from pandas import Series

from psdmpy.models.result.grid.connector import ConnectorResult, ConnectorsResult
from psdmpy.models.result.power import PQResult

if TYPE_CHECKING:
    from psdmpy.models.gwr import GridWithResults


@dataclass(frozen=True)
class Transformer2WResult(ConnectorResult):
    @property
    def tap_pos(self) -> Series:
        return self.data["tap_pos"]

    def calc_apparent_power_gwr(self, gwr: GridWithResults, side="hv"):
        transformer = gwr.grid.raw_grid.transformers_2_w.subset(self.input_model)
        node_uuid = transformer.node_a if side == "hv" else transformer.node_b
        node_uuid = node_uuid.to_list()[0]
        v_rated = gwr.grid.raw_grid.nodes.v_rated.loc[node_uuid]
        node_res = gwr.results.nodes[node_uuid]
        node_side = self._get_node_for_side(side)
        return self.calc_apparent_power(
            node_res, voltage_level_kv=v_rated, node=node_side
        )

    def calc_active_power_gwr(self, gwr: GridWithResults, side="hv"):
        s = self.calc_apparent_power_gwr(gwr, side)
        return s.apply(lambda x: np.real(x)).rename("p")

    def calc_reactive_power_gwr(self, gwr: GridWithResults, side="hv"):
        s = self.calc_apparent_power_gwr(gwr, side)
        return s.apply(lambda x: np.imag(x)).rename("q")

    def calc_transformer_utilisation(self, gwr: GridWithResults, side="hv"):
        transformer = gwr.grid.raw_grid.transformers_2_w.subset(self.input_model)
        s_rated = transformer.s_rated.to_list()[0]
        return (
            self.calc_apparent_power_gwr(gwr, side).apply(lambda x: np.abs(x)) / s_rated
        )

    def to_pq_result(self, gwr: GridWithResults, side="hv"):
        p = self.calc_active_power_gwr(gwr, side)
        q = self.calc_reactive_power_gwr(gwr, side)
        data = pd.concat([p, q], axis=1)
        return PQResult(self.entity_type, self.input_model, self.name, data)

    @staticmethod
    def attributes() -> List[str]:
        return ConnectorResult.attributes() + ["tap_pos"]

    @staticmethod
    def _get_node_for_side(side):
        if side == "hv":
            return "a"
        elif side == "lv":
            return "b"
        else:
            raise ValueError('Side has to be either "hv" or "lv"')


@dataclass(frozen=True)
class Transformers2WResult(ConnectorsResult):
    entities: dict[str, Transformer2WResult]



import math
from datetime import datetime
from typing import Optional, Union

import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns
from matplotlib.axes import Axes

from psdmpy.models.gwr import GridWithResults
from psdmpy.models.result.grid.enhanced_node import EnhancedNodesResult
from psdmpy.models.result.grid.node import NodeResult, NodesResult
from psdmpy.plots.common.line_plot import ax_plot_time_series
from psdmpy.plots.common.utils import (
    COLOR_PALETTE,
    FIGSIZE,
    LABEL_PAD,
    set_subplot_title,
    set_title,
    set_xlabels_rotated,
    set_ylabel,
)


def plot_all_v_mag_branch_violin(
    gwr: GridWithResults,
):
    """
    Plots violin plots for all nodes across all branches.

    Args:
        gwr: GridWithResults object

    Returns:
        fig, axes
    """
    branches = gwr.grid.raw_grid.get_branches()
    nodes_res = gwr.results.nodes
    width, height = FIGSIZE
    height = height * len(branches)
    fig, axes = plt.subplots(nrows=len(branches), figsize=(width, height))
    for i, branch in enumerate(branches):
        ax_plot_v_mags_violin(axes[i], nodes_res, branch)
        set_subplot_title(axes[i], f"Voltages along Branch {i+1}")
    plt.tight_layout()

    return fig, axes


def plot_v_mags_violin(
    nodes_res: Union[NodesResult, EnhancedNodesResult],
    nodes: Optional[list[str]] = None,
):
    """
    Plots violin plots for all given nodes .

    Args:
        nodes_res: NodesResult or EnhancedNodesResult object
        nodes: Optional list of node uuids that should be plotted. Order is preserved.

    Returns:
        fig, ax
    """
    fig, ax = plt.subplots(figsize=FIGSIZE)
    ax_plot_v_mags_violin(ax, nodes_res, nodes)
    set_title(ax, "Voltage Magnitudes along Branch")
    return fig, ax


def ax_plot_v_mags_violin(
    ax: Axes,
    nodes_res: Union[NodesResult, EnhancedNodesResult],
    nodes: Optional[list[str]],  # branches can be found by GridContainer.get_branches()
):
    """
    Plots violin plots for given nodes. If no nodes are passed all nodes are plotted.

    Args:
        ax: Axes object
        nodes_res: NodesResult or EnhancedNodesResult object
        nodes: Optional list of node uuids that should be plotted. Order is preserved.
    """

    if nodes:
        # get v_mag in listed sequence
        v_mag = nodes_res.subset(nodes).v_mag.reindex(columns=nodes)
    else:
        v_mag = nodes_res.v_mag

    sns.violinplot(v_mag, showmedians=True, ax=ax, linewidth=0.5, palette=COLOR_PALETTE)

    # set labels
    uuid_to_id = nodes_res.uuid_to_id_map()
    x_labels = v_mag.columns.map(lambda uuid: uuid_to_id[uuid])
    set_xlabels_rotated(ax, x_labels, ha="right")
    set_ylabel(ax, "Voltage magnitude in pu")
    _ = ax.set_xticklabels(x_labels, rotation=45, ha="right")


def plot_v_mag_branch(
    nodes_res: Union[NodesResult, EnhancedNodesResult],
    branch: list[str],  # branches can be found by GridContainer.get_branches()
    time: datetime,
    in_kw: bool = False,
):
    """
    Plots voltage magnitudes of nodes along the given branch.

    Args:
        nodes_res: NodesResult or EnhancedNodesResult object
        branch: list of node uuids that form the branch
        time: datetime for which the voltage magnitudes should be plotted
        in_kw: if True, power is plotted in kW instead of MW

    Returns:
        fig, ax
    """
    fig, ax = plt.subplots(figsize=(10, 5))
    ax_plot_v_mag_branch(ax, nodes_res, branch, time, fig=fig, in_kw=in_kw)
    set_title(ax, "Voltage Magnitudes along Branch")
    return fig, ax


def ax_plot_v_mag_branch(
    ax: Axes,
    nodes_res: Union[NodeResult, EnhancedNodesResult],
    branch: list[str],
    time: datetime,
    fig: Optional[plt.Figure] = None,  # used to plot colorbar
    in_kw: bool = False,
):
    """
    Plots voltage magnitudes of nodes along the given branch.

    Args:
        ax: Axes object
        nodes_res: NodesResult or EnhancedNodesResult object
        branch: list of node uuids that form the branch
        time: datetime for which the voltage magnitudes should be plotted
        fig: Optional Figure object which is used to plot the colorbar
        in_kw: if True, power is plotted in kW instead of MW
    """

    with_power = isinstance(nodes_res, EnhancedNodesResult)

    v_mags = []
    x_ticks = []
    p_s = []

    for node in branch:
        node_res = nodes_res[node]
        res = node_res[time]
        v_mag = res.v_mag.iloc[0]

        # happens for auxiliary node at open switch
        if math.isnan(v_mag):
            continue
        v_mags.append(v_mag)

        if with_power:
            p_s.append(res.p.iloc[0])
        x_ticks.append(node_res.name)

    ax.plot(x_ticks, v_mags)
    ax.set_xticks(np.arange(len(x_ticks)))  # xtick locations
    ax.set_xticklabels(x_ticks, rotation=45, ha="right")  # set xtick labels
    set_ylabel(ax, "Voltage Magnitude in p.u.")

    if with_power:
        # plot aggregated node power indicator
        cmap = sns.color_palette("coolwarm", as_cmap=True)
        if in_kw:
            p_s = [p * 1000 for p in p_s]
        sc = ax.scatter(x_ticks, v_mags, c=p_s, cmap=cmap, edgecolor="none", zorder=10)
        if fig:
            cbar = fig.colorbar(sc, ax=ax)
            unit = "kW" if in_kw else "MW"
            cbar.set_label(f"Nodal Power in {unit}", labelpad=LABEL_PAD)


def plot_v_mag(
    res: NodeResult,
    resolution: str,
    title: Optional[str] = None,
    fill_from_index: bool = False,
    fill_between=None,
    set_x_label=True,
    **kwargs,
):
    if title is None:
        title = f"Voltage Magnitude at Node: {res.name}"
    fig, ax = plt.subplots(figsize=FIGSIZE)
    ax_plot_node_v_mag(
        ax,
        res,
        resolution,
        fill_from_index=fill_from_index,
        fill_between=fill_between,
        set_x_label=set_x_label,
        **kwargs,
    )
    set_title(ax, title)
    return fig, ax


def plot_v_ang(
    res: NodeResult,
    resolution: str,
    title: Optional[str] = None,
    hourly_mean: bool = False,
    fill_from_index: bool = False,
    fill_between=None,
    set_x_label=True,
    **kwargs,
):
    if title is None:
        title = f"Voltage Angle at Node: {res.name}"
    fig, ax = plt.subplots(figsize=FIGSIZE)
    ax_plot_node_v_ang(
        ax,
        res,
        resolution,
        hourly_mean=hourly_mean,
        fill_from_index=fill_from_index,
        fill_between=fill_between,
        set_x_label=set_x_label,
        **kwargs,
    )
    set_title(ax, title)
    return fig, ax


def ax_plot_node_v_mag(
    ax: Axes,
    res: NodeResult,
    resolution: str,
    fill_from_index: bool = False,
    fill_between=None,
    set_x_label=True,
    **kwargs,
):
    if len(res.v_mag) == 0:
        raise ValueError("Voltage magnitude time series is empty. No data to plot")

    ax = ax_plot_time_series(
        ax,
        res.v_mag,
        res.entity_type,
        resolution,
        fill_from_index=fill_from_index,
        fill_between=fill_between,
        set_x_label=set_x_label,
        **kwargs,
    )
    ax.set_ylabel("Voltage Magnitude in p.u.")


def ax_plot_node_v_ang(
    ax: Axes,
    res: NodeResult,
    resolution: str,
    fill_from_index: bool = False,
    fill_between=None,
    set_x_label=True,
    **kwargs,
):
    if len(res.v_ang) == 0:
        raise ValueError("Voltage angle time series is empty. No data to plot")

    ax = ax_plot_time_series(
        ax,
        res.v_ang,
        res.entity_type,
        resolution,
        fill_from_index=fill_from_index,
        fill_between=fill_between,
        set_x_label=set_x_label,
        **kwargs,
    )
    ax.set_ylabel("Voltage Angle in deg")



import pandas as pd

from psdmpy.models.input.container.raw_grid import RawGridContainer


def grid_structure_summary(grid: RawGridContainer):
    line_length_sum = grid.lines.aggregated_line_length()
    return """
    Grid Structure Summary:
    -----------------------
    {nr_2w_transformers} two-winding transformers
    {nr_nodes} nodes
    {nr_lines} lines
    {line_length_sum} km aggregated line length
    """.format(
        nr_2w_transformers=len(grid.transformers_2_w.data),
        nr_nodes=len(grid.nodes.data),
        nr_lines=len(grid.lines.data),
        line_length_sum=line_length_sum,
    )


def branching_index(grid: RawGridContainer):
    """
    Calculates the average amount of branches connected to the nodes of the grid.

    Args:
        grid: The grid to calculate the branching index for.

    Returns:
        The branching index of the grid.
    """
    # all connections to nodes
    connections = pd.concat((grid.lines.node_a, grid.lines.node_b))
    # TODO: add switch evaluation
    # group connections by node uuid and calculate their amount
    nodal_connections = connections.groupby(connections).apply(
        lambda node_grp: len(node_grp)
    )
    return nodal_connections.sum() / len(nodal_connections)



from dataclasses import dataclass
from typing import List

from psdmpy.models.enums import EntitiesEnum, RawGridElementsEnum
from psdmpy.models.input.connector.connector import Connector


@dataclass(frozen=True)
class Switches(Connector):
    @property
    def closed(self):
        return self.data["closed"]

    def get_closed(self) -> "Switches":
        return self.subset(self.data.query("closed == True").index)

    def get_opened(self) -> "Switches":
        return self.subset(self.data.query("closed == False").index)

    @staticmethod
    def attributes() -> list[str]:
        # ignore parallel devices since this is always 1
        # also we cannot read the attribute in the PSDM at the moment
        return [
            attr
            for attr in Connector.attributes() + ["closed"]
            if attr != "parallel_devices"
        ]

    @staticmethod
    def get_enum() -> EntitiesEnum:
        return RawGridElementsEnum.SWITCH

    @staticmethod
    def bool_attributes() -> List[str]:
        return Connector.bool_attributes() + ["closed"]



from __future__ import annotations

import copy
import json
import logging
import os
from abc import ABC, abstractmethod
from dataclasses import dataclass, replace
from pathlib import Path
from typing import TYPE_CHECKING, List, Tuple, TypeVar, Union

import pandas as pd
from pandas import DataFrame, Series

from psdmpy.errors import ComparisonError
from psdmpy.io import utils
from psdmpy.io.utils import bool_converter, df_to_csv, read_csv
from psdmpy.models.enums import (
    EntitiesEnum,
    RawGridElementsEnum,
    SystemParticipantsEnum,
)
from psdmpy.models.input.participant.charging import parse_evcs_type_info
from psdmpy.processing.dataframe import compare_dfs

if TYPE_CHECKING:
    from psdmpy.models.input.node import Nodes

EntityType = TypeVar("EntityType", bound="Entities")


@dataclass(frozen=True)
class Entities(ABC):
    """
    Entities is the abstract base class for all input entity models.
    At it's core all data can be retrieved by accessing the data attribute which returns a pandas DataFrame.
    Every row corresponds to one entity, indexed with their uuid.
    Attribute columns can be accessed by their respectively named property methods.

    Args:
        data: The data of the entities.
    """

    data: DataFrame

    def __len__(self):
        return len(self.data)

    def __repr__(self):
        return self.data

    def __contains__(self, uuid: str):
        return uuid in self.data.index

    def __add__(self: EntityType, other: EntityType) -> EntityType:
        """
        Concatenates two Entities instances.

        Args:
            other: The other Entities instance to concatenate with.

        Returns:
            The concatenated Entities instance.
        """

        if not isinstance(other, type(self)):
            raise TypeError(
                f"The two Entities instances must be of the same type. Received {type(self)} and {type(other)}"
            )

        columns_diff = set(self.data.columns).symmetric_difference(other.data.columns)
        if len(columns_diff) > 0 and columns_diff.issubset(self.attributes()):
            logging.warning(
                "The two Entities instances have different columns: %s", columns_diff
            )
        return type(self)(pd.concat([self.data, other.data]))

    def __sub__(self: EntityType, other: Union[EntityType, List[str]]) -> EntityType:
        """
        Subtacts the entities with the given uuids from the Entities instance.

        Args:
            other: The other Entities instance or a list of uuids to subtract.

        Returns:
            A new Entities instance with the uuids removed.
        """

        if isinstance(other, Entities):
            indices_to_remove = other.data.index
        elif isinstance(other, list) and all(isinstance(index, str) for index in other):
            indices_to_remove = other
        else:
            raise TypeError("other must be an Entities instance or a list of strings")

        if not set(indices_to_remove).issubset(self.data.index):
            raise ValueError(
                f"All indices to remove must exist in the current Entities instance: {set(indices_to_remove) - set(self.data.index)}"
            )

        return type(self)(self.data.drop(indices_to_remove))

    @property
    def uuid(self) -> Series:
        """
        Returns: The uuids of the entities.
        """
        return self.data.index

    @property
    def id(self) -> Series:
        """
        Returns: The ids of the entities.
        """
        return self.data["id"]

    @property
    def operates_from(self) -> Series:
        """
        Returns: The operaton start time of the entities.
        """
        return self.data["operates_from"]

    @property
    def operates_until(self) -> Series:
        """
        Returns: The operaton end time of the entities.
        """
        return self.data["operates_until"]

    @property
    def operator(self) -> Series:
        """
        Returns: The operator of the entities.
        """
        return self.data["operator"]

    @property
    @abstractmethod
    def node(self) -> Series:
        """
        Returns: The nodes to which the entities are connected.
        """
        pass

    def get(self, uuid: str) -> Series:
        """
        Returns the entity information of the entitiy with the given uuid.

        Args:
            uuid: The uuid of the entity.

        Returns:
            Row (Series) of the corresponding entity information.
        """
        return self.data.loc[uuid]

    def subset(self: EntityType, uuids: Union[list[str], set[str], str]) -> EntityType:
        """
        Creates a subset of the Entities instance with the given uuids.

        Args:
            uuids: The uuids to subset.

        Returns:
            A new instance with the subset of entities.
        """
        if isinstance(uuids, str):
            uuids = [uuids]
        elif isinstance(uuids, set):
            uuids = list(uuids)
        try:
            return type(self)(self.data.loc[uuids])
        except KeyError as e:
            not_found = set(uuids) - set(self.data.index)
            raise KeyError(
                f"uuids must be a subset of the current Entities instance. The following uuids couldn't be found: {not_found}"
            ) from e

    def subset_id(self: EntityType, ids: Union[list[str], set[str], str]) -> EntityType:
        """
        Creates a subset of the Entities instance with the given ids.

        Args:
            ids: The ids to subset.

        Returns:
            A new instance with the subset of entities.
        """
        if isinstance(ids, str):
            ids = [ids]
        elif isinstance(ids, set):
            ids = list(ids)
        return type(self)(self.data[self.data["id"].isin(ids)])

    def subset_split(
        self: EntityType, uuids: Union[list[str], set[str], str]
    ) -> Tuple[EntityType, EntityType]:
        """
        Returns the subset of entities as well as the remaining instances.

        Args:
            uuids: The uuids to subset.

        Returns:
            A tuple of the subset and the remaining entities.
        """
        rmd = set(self.uuid) - set(uuids)
        return self.subset(uuids), self.subset(list(rmd))

    def filter_by_nodes(
        self: EntityType, nodes: Union[str, list[str], set[str]]
    ) -> EntityType:
        """
        Filters for all entities which are connected to one of th given nodes.

        Args:
            nodes: The nodes to filter by, represented by their uuids.

        Returns:
            A new Entities instance with the filtered entities.
        """
        if isinstance(nodes, str):
            nodes = [nodes]
        data = self.data[self.node.isin(nodes)]
        return type(self)(data)

    def find_nodes(self, nodes: Nodes) -> Nodes:
        return nodes.subset(self.node)

    def to_csv(self, path: str, mkdirs=True, delimiter: str = ","):
        # local import to avoid circular imports
        from psdmpy.models.input.container.mixins import HasTypeMixin

        if isinstance(self, HasTypeMixin):
            HasTypeMixin.to_csv(self, path, mkdirs, delimiter)
        else:
            df_to_csv(
                self.data,
                path,
                self.get_enum().get_csv_input_file_name(),
                mkdirs=mkdirs,
                delimiter=delimiter,
            )

    def compare(self, other) -> None:
        """
        Compares the current Entities instance with another Entities instance.

        Args:
            other: The other Entities instance to compare with.

        Returns:
            None
        """
        if not isinstance(other, type(self)):
            raise ComparisonError(
                f"Type of self {type(self)} != type of other {type(other)}"
            )

        try:
            compare_dfs(self.data, other.data)
        except AssertionError as e:
            raise ComparisonError(
                f"{self.get_enum().get_plot_name()} entities are not equal.",
                errors=[(type(self), str(e))],
            )

    def copy(
        self: EntityType,
        deep=True,
        **changes,
    ) -> EntityType:
        """
        Creates a copy of the current Entities instance.
        By default does a deep copy of all data and replaces the given changes.
        When deep is false, only the references to the data of the non-changed attribtues are copied.

        Args:
            deep: Whether to do a deep copy of the data.
            **changes: The changes to apply to the copy.

        Returns:
            The copy of the current Entities instance.
        """
        to_copy = copy.deepcopy(self) if deep else self
        return replace(to_copy, **changes)

    @classmethod
    def from_csv(cls: EntityType, path: str, delimiter: str) -> EntityType:
        """
        Reads the entity data from a csv file.

        Args:
            path: The path to the csv file.
            delimiter: The delimiter of the csv file.

        Returns:
           The corresponding entities object.
        """
        return cls._from_csv(path, delimiter, cls.get_enum())

    @classmethod
    def _from_csv(cls, path: str, delimiter: str, entity: EntitiesEnum):
        file_path = utils.get_file_path(path, entity.get_csv_input_file_name())
        if os.path.exists(file_path):
            return cls(cls._data_from_csv(entity, path, delimiter))
        else:
            logging.debug(
                "There is no file named: "
                + str(file_path)
                + ". No "
                + entity.value
                + " entities are loaded."
            )
            return cls.create_empty()

    @classmethod
    def _data_from_csv(
        cls, entity: EntitiesEnum, path: str | Path, delimiter: str
    ) -> DataFrame:
        data = read_csv(path, entity.get_csv_input_file_name(), delimiter)

        bool_cols = cls.bool_attributes()
        for col in bool_cols:
            try:
                data[col] = data[col].apply(lambda x: bool_converter(x))
                data[col] = data[col].astype(bool)
            except ValueError as e:
                logging.error(
                    f"Could not convert column {col} to bool. "
                    f"Please check the values in the csv file. "
                    f"Error: {e}"
                )

        if entity.has_type():
            type_data = read_csv(path, entity.get_type_file_name(), delimiter)
            data = (
                data.merge(
                    type_data, left_on="type", right_on="uuid", suffixes=("", "_type")
                )
                .rename(columns={"id_type": "type_id", "uuid_type": "type_uuid"})
                .drop(columns=["type"])
            )
        # special data transformations
        match entity:
            # for raw grid elements
            # ---------------------
            case RawGridElementsEnum.NODE:
                data["longitude"] = data["geo_position"].apply(
                    lambda geo_json: json.loads(geo_json)["coordinates"][0]
                )
                data["latitude"] = data["geo_position"].apply(
                    lambda geo_json: json.loads(geo_json)["coordinates"][1]
                )

            # for system participants
            # -----------------------
            case SystemParticipantsEnum.ENERGY_MANAGEMENT:
                data["connected_assets"] = data["connected_assets"].apply(
                    lambda x: x.split(" ")
                )
            case SystemParticipantsEnum.EV_CHARGING_STATION:
                type_data = data["type"].apply(
                    lambda type_str: parse_evcs_type_info(type_str)
                )
                data = pd.concat([data, type_data])

        try:
            return data.set_index("uuid")
        except KeyError as e:
            raise KeyError(
                "Column 'uuid' not found. This might be due to wrong csv delimiter!", e
            )

    @classmethod
    def create_empty(cls: EntityType) -> EntityType:
        """
        Creates an empty instance of the corresponding entity class.
        """
        data = pd.DataFrame(columns=cls.attributes())
        return cls(data)

    @staticmethod
    @abstractmethod
    def get_enum() -> EntitiesEnum:
        """
        Returns the corresponding entity enum value.
        """
        pass

    @staticmethod
    def attributes() -> List[str]:
        """
        Method that should hold all attributes field (transformed to snake_case and case-sensitive)
        of the corresponding PSDM entity
        :return:
        """
        return ["id", "operates_from", "operates_until", "operator"]

    @staticmethod
    def bool_attributes() -> List[str]:
        """
        Method that should hold all boolean attributes field (transformed to snake_case and case-sensitive)
        of the corresponding PSDM entity. It is mainly used for type casting of the data frame column.
        :return:
        """
        return []



import matplotlib.pyplot as plt
from matplotlib import gridspec

from psdmpy.models.gwr import GridWithResults
from psdmpy.plots.common.utils import FIGSIZE
from psdmpy.plots.results.power_plot import ax_plot_nodal_ps_violin
from psdmpy.plots.results.voltage_plot import ax_plot_v_mags_violin


def voltage_power_along_branches_violin(gwr: GridWithResults):
    branches = gwr.grid.raw_grid.get_branches()
    nodes_res = gwr.build_enhanced_nodes_result()
    width, height = FIGSIZE
    height = height * len(branches) * 2
    fig = plt.figure(figsize=(width, height))
    outer_grid = gridspec.GridSpec(len(branches), 1, wspace=0.2, hspace=0.3)
    inner_grids = []
    for i, branch in enumerate(branches):
        inner_grid = gridspec.GridSpecFromSubplotSpec(
            2, 1, subplot_spec=outer_grid[i], wspace=0.1, hspace=0.6
        )
        inner_grids.append(inner_grid)
        ax_v = plt.Subplot(fig, inner_grid[0])
        ax_plot_v_mags_violin(ax_v, nodes_res, branch)
        ax_v.set_title(f"Voltage and Active Power for Branch {i}")
        fig.add_subplot(ax_v)

        ax_p = plt.Subplot(fig, inner_grid[1])
        ax_plot_nodal_ps_violin(ax_p, nodes_res, branch)
        fig.add_subplot(ax_p)

    plt.tight_layout()
    plt.show()

    return fig, outer_grid, inner_grids



import concurrent.futures
from dataclasses import dataclass
from datetime import datetime
from typing import Optional, Union

from psdmpy.io.utils import check_filter
from psdmpy.models.enums import RawGridElementsEnum
from psdmpy.models.input.container.grid import GridContainer
from psdmpy.models.input.container.mixins import ContainerMixin
from psdmpy.models.result.container.participants import (
    ParticipantsResultContainer,
)
from psdmpy.models.result.grid.connector import ConnectorsResult
from psdmpy.models.result.grid.node import NodesResult
from psdmpy.models.result.grid.transformer import Transformers2WResult


@dataclass(frozen=True)
class GridResultContainer(ContainerMixin):
    name: str
    nodes: NodesResult
    lines: ConnectorsResult
    transformers_2w: ConnectorsResult
    participants: ParticipantsResultContainer

    def __len__(self):
        return (
            len(self.nodes)
            + len(self.lines)
            + len(self.transformers_2w)
            + len(self.participants)
        )

    # todo: implement slicing
    def __getitem__(self, slice_val):
        raise NotImplementedError

    def to_list(self, include_empty: bool = False) -> list:
        res = [self.nodes, self.lines, self.transformers_2w, self.participants]
        return res if include_empty else [r for r in res if r]

    def uuids(self) -> set[str]:
        return set(self.nodes.entities.keys())

    def filter_by_date_time(self, time: Union[datetime, list[datetime]]):
        return GridResultContainer(
            self.name,
            self.nodes.filter_by_date_time(time),
            self.lines.filter_by_date_time(time),
            self.transformers_2w.filter_by_date_time(time),
            self.participants.filter_by_date_time(time),
        )

    def filter_for_time_interval(self, start: datetime, end: datetime):
        return GridResultContainer(
            self.name,
            self.nodes.filter_for_time_interval(start, end),
            self.lines.filter_for_time_interval(start, end),
            self.transformers_2w.filter_for_time_interval(start, end),
            self.participants.filter_for_time_interval(start, end),
        )

    @classmethod
    def from_csv(
        cls,
        name: str,
        simulation_data_path: str,
        delimiter: str,
        simulation_end: Optional[datetime] = None,
        grid_container: Optional[GridContainer] = None,
        filter_start: Optional[datetime] = None,
        filter_end: Optional[datetime] = None,
    ):
        check_filter(filter_start, filter_end)

        with concurrent.futures.ProcessPoolExecutor() as executor:
            nodes_future = executor.submit(
                NodesResult.from_csv,
                simulation_data_path,
                delimiter,
                simulation_end,
                grid_container.raw_grid.nodes if grid_container else None,
                filter_start,
                filter_end,
            )
            transformers_2_w_future = executor.submit(
                Transformers2WResult.from_csv,
                RawGridElementsEnum.TRANSFORMER_2_W,
                simulation_data_path,
                delimiter,
                simulation_end,
                grid_container.raw_grid.transformers_2_w if grid_container else None,
                filter_start,
                filter_end,
            )
            lines_future = executor.submit(
                ConnectorsResult.from_csv,
                RawGridElementsEnum.LINE,
                simulation_data_path,
                delimiter,
                simulation_end,
                grid_container.raw_grid.lines if grid_container else None,
                filter_start,
                filter_end,
            )

            nodes = nodes_future.result()
            transformers_2_w = transformers_2_w_future.result()
            lines = lines_future.result()

        if simulation_end is None:
            if len(nodes.entities) == 0:
                raise ValueError(
                    "Can't determine simulation end time automatically. No node results to base it on. Please configure 'simulation_end' manually."
                )
            some_node_res = next(iter(nodes.entities.values()))
            simulation_end = some_node_res.data.index.max()

        participants = ParticipantsResultContainer.from_csv(
            simulation_data_path,
            delimiter,
            simulation_end,
            grid_container=grid_container,
            filter_start=filter_start,
            filter_end=filter_end,
        )

        return cls(name, nodes, lines, transformers_2_w, participants)

    @classmethod
    def create_empty(cls):
        return cls(
            name="Empty Container",
            nodes=NodesResult.create_empty(),
            lines=ConnectorsResult.create_empty(RawGridElementsEnum.LINE),
            transformers_2w=ConnectorsResult.create_empty(
                RawGridElementsEnum.TRANSFORMER_2_W
            ),
            participants=ParticipantsResultContainer.create_empty(),
        )



# Thermal Bus
from uuid import uuid4

import pandas as pd

from psdmpy.models.input.create.utils import create_data
from psdmpy.models.input.thermal.bus import ThermalBus
from psdmpy.models.input.thermal.house import ThermalHouse


def create_thermal_bus_data(
    id,
    uuid=None,
    operates_from=None,
    operates_until=None,
    operator=None,
):
    if not uuid:
        uuid = str(uuid4())
    return pd.Series(
        {
            "id": id,
            "operates_from": operates_from,
            "operates_until": operates_until,
            "operator": operator,
        }
    ).rename(uuid)


def create_thermal_busses(data_dict):
    return ThermalBus(create_data(data_dict, create_thermal_bus_data))


def create_thermal_house_data(
    id,
    thermal_bus,
    eth_losses,
    eth_capa,
    target_temperature=20,
    upper_temperature_limit=22,
    lower_temperature_limit=18,
    uuid=None,
    operates_from=None,
    operates_until=None,
    operator=None,
):
    if not uuid:
        uuid = str(uuid4())
    return pd.Series(
        {
            "id": id,
            "thermal_bus": thermal_bus,
            "operates_from": operates_from,
            "operates_until": operates_until,
            "operator": operator,
            "eth_losses": eth_losses,
            "eth_capa": eth_capa,
            "target_temperature": target_temperature,
            "upper_temperature_limit": upper_temperature_limit,
            "lower_temperature_limit": lower_temperature_limit,
        }
    ).rename(uuid)


def create_thermal_houses(data_dict):
    return ThermalHouse(create_data(data_dict, create_thermal_house_data))



import os.path
from typing import Dict, NewType, Tuple

import seaborn as sns
from matplotlib import dates as mdates
from matplotlib.axes import Axes
from matplotlib.figure import Figure
from pandas import Series

from psdmpy.models.enums import (
    EntitiesEnum,
    RawGridElementsEnum,
    SystemParticipantsEnum,
)
from psdmpy.processing.series import hourly_mean_resample

# === COLORS ===

COLOR_PALETTE = sns.color_palette()

BLUE = COLOR_PALETTE[0]
ORANGE = COLOR_PALETTE[1]
GREEN = COLOR_PALETTE[2]
RED = COLOR_PALETTE[3]
PURPLE = COLOR_PALETTE[4]
BROWN = COLOR_PALETTE[5]
PINK = COLOR_PALETTE[6]
GREY = COLOR_PALETTE[7]
YELLOW = COLOR_PALETTE[8]
LIGHT_BLUE = COLOR_PALETTE[9]

# === COLOR MATCHING ===

NODE_COLOR = RED
TRANSFORMER_COLOR = RED
LINE_COLOR = BLUE
LOAD_COLOR = BLUE
PV_COLOR = GREEN
BS_COLOR = ORANGE
HP_COLOR = PURPLE
EVCS_COLOR = YELLOW
RESIDUAL_LOAD_COLOR = LIGHT_BLUE
UNKNOWN_COLOR = GREY
FLEX_MAX = BLUE
FLEX_MIN = GREEN
FLEX_REF = YELLOW

RGB = NewType("RGB", Tuple[float, float, float])


def get_label_and_color(sp_type: EntitiesEnum) -> Tuple[str, RGB]:
    match sp_type:
        case RawGridElementsEnum.NODE:
            return "Node", NODE_COLOR
        case RawGridElementsEnum.TRANSFORMER_2_W:
            return "Transformer", TRANSFORMER_COLOR
        case RawGridElementsEnum.LINE:
            return "Line", LINE_COLOR
        case SystemParticipantsEnum.LOAD:
            return "Load", LOAD_COLOR
        case SystemParticipantsEnum.PHOTOVOLTAIC_POWER_PLANT:
            return "PV", PV_COLOR
        case SystemParticipantsEnum.STORAGE:
            return "Battery", BS_COLOR
        case SystemParticipantsEnum.EV_CHARGING_STATION:
            return "EV Charging Station", EVCS_COLOR
        case SystemParticipantsEnum.ELECTRIC_VEHICLE:
            return "Electric Vehicle", EVCS_COLOR
        case SystemParticipantsEnum.HEAT_PUMP:
            return "Heat Pump", HP_COLOR
        case SystemParticipantsEnum.ENERGY_MANAGEMENT:
            return "Energy Management", LOAD_COLOR
        case SystemParticipantsEnum.PARTICIPANTS_SUM:
            return "Participants Sum", LOAD_COLOR
        case SystemParticipantsEnum.PRIMARY_DATA:
            return "Primary Data", LOAD_COLOR
        case _:
            return sp_type.value, UNKNOWN_COLOR


def get_label_and_color_dict(sp_type: EntitiesEnum):
    label, color = get_label_and_color(sp_type)
    return {"label": label, "color": color}


# === FIGURE DEFAULTS ===

FIGSIZE = (12, 4)
FIGSIZE_WIDE = (15, 4)
SUBPLOTS_PADDING = 10

# === FONT DEFAULTS ===

TITLE_FONT_SIZE = 14

# === SPACING ===

TITLE_PAD = 15
LABEL_PAD = 10

# === SHADING ===

FILL_ALPHA = 0.2


def save_fig(figure: Figure, path: str, file_name: str, format="svg"):
    figure.savefig(os.path.join(path, file_name), bbox_inches="tight", format=format)


def plot_resample(res: Series, hourly_mean: bool):
    return (
        res.resample("60S").ffill()
        if not hourly_mean
        # todo: check the resample 60s
        else hourly_mean_resample(res).resample("60S").ffill()
    )


def set_date_format_and_label(ax: Axes, resolution: str):
    date_format, x_label = _date_format_and_x_label(resolution)
    ax.set_xlabel(x_label)
    if resolution == "y":
        # set x labels for every month
        locator = mdates.MonthLocator()
        ax.get_xaxis().set_major_locator(locator)
    if resolution == "m":
        locator = mdates.WeekdayLocator()
        ax.get_xaxis().set_major_locator(locator)
    ax.xaxis.set_major_formatter(mdates.DateFormatter(date_format))
    ax.xaxis.set_minor_formatter(mdates.DateFormatter(date_format))


def _date_format_and_x_label(resolution: str):
    if resolution == "d":
        return "%H:%M", "Time in h"
    elif resolution == "w":
        return "%a", "Day of Week"
    elif resolution == "m":
        return "%x", "Day of Month"
    elif resolution == "y":
        return "%b", "Month"


def set_suptitle(fig: Figure, title: str):
    fig.suptitle(title, fontsize=TITLE_FONT_SIZE)


def set_title(ax: Axes, title: str):
    ax.set_title(title, pad=TITLE_PAD, fontsize=TITLE_FONT_SIZE)


def set_subplot_title(ax: Axes, title: str):
    ax.set_title(title, pad=TITLE_PAD)


def set_ylabel(ax: Axes, ylabel: str):
    ax.set_ylabel(ylabel, labelpad=LABEL_PAD)


def set_xlabels_rotated(
    ax: Axes, labels: list[str], rotation: int = 45, ha: str = "center"
):
    ax.set_xticks(range(1, len(labels) + 1))
    ax.set_xticklabels(labels, rotation=rotation, ha=ha)


def legend_with_distinct_labels(ax: Axes):
    handles, labels = ax.get_legend_handles_labels()
    by_label = dict(zip(labels, handles))
    ax.legend(by_label.values(), by_label.keys())


def rgb_to_hex(rgb):
    return "#%02x%02x%02x" % (int(rgb[0] * 255), int(rgb[1] * 255), int(rgb[2] * 255))


def add_to_kwargs_if_not_exist(kwargs: Dict, to_add: Dict):
    for key, value in to_add.items():
        if key not in kwargs:
            kwargs[key] = value
    return kwargs


def ax_plot_secondary_axis(
    ax: Axes, series: Series, ax_y_label: str, show_secondary_grid_lines: bool = False
):
    ax2 = ax.twinx()
    (line_plot,) = ax2.plot(series)
    line_plot.set_visible(False)
    ax2.set_zorder(ax.get_zorder() - 1)
    ax2.set_ylabel(ax_y_label, labelpad=10)
    if show_secondary_grid_lines:
        # Make the background of the original y-axis transparent so the gridlines are visible
        ax.patch.set_visible(False)
    return ax2
